---@alias ChallengeType integer
CHALLENGE_TIMED = 1 << 0
CHALLENGE_TOXIC_GAS = 1 << 1
CHALLENGE_COSMIC_CLONE = 1 << 2
CHALLENGE_COIN_REQUIREMENT = 1 << 3
CHALLENGE_CAPLESS = 1 << 4
CHALLENGE_RISING_LAVA = 1 << 5
CHALLENGE_DAREDEVIL = 1 << 6
CHALLENGE_ACID_RAIN = 1 << 7
CHALLENGE_QUICKSAND_PIT = 1 << 8
CHALLENGE_PVP_RACE = 1 << 9
CHALLENGE_GROUNDLESS = 1 << 10
CHALLENGE_DARKNESS = 1 << 11
CHALLENGE_ABC = 1 << 12
CHALLENGE_NO_BLJS = 1 << 13

E_MODEL_NOVA_STAR = smlua_model_util_get_id("nova_star_geo")
E_MODEL_NOVA_STAR_TRANSPARENT = smlua_model_util_get_id("nova_star_transparent_geo")
E_MODEL_NOVA_STAR_SHINE = smlua_model_util_get_id("nova_shine_geo")
E_MODEL_SQUISHY_STAR = smlua_model_util_get_id("squishy_star_geo")

E_MODEL_ZTAR = smlua_model_util_get_id("ztar_geo")
E_MODEL_ZTAR_TRANSPARENT = smlua_model_util_get_id("transparent_ztar_geo")
E_MODEL_ZTAR_SHINE = smlua_model_util_get_id("ztar_shine_geo")

E_MODEL_COSMIC_STAR = smlua_model_util_get_id("cosmic_star_geo")
E_MODEL_COSMIC_STAR_TRANSPARENT = E_MODEL_TRANSPARENT_STAR
E_MODEL_COSMIC_STAR_SHINE = smlua_model_util_get_id("cosmic_star_shine_geo")

gNovaStarModels = {
    ["vanilla"] = {
        ["vanilla moveset"] = {normal = E_MODEL_ZTAR, transparent = E_MODEL_ZTAR_TRANSPARENT, shine = E_MODEL_ZTAR_SHINE},
        ["omm rebirth 1.4"] = {normal = E_MODEL_NOVA_STAR, transparent = E_MODEL_NOVA_STAR_TRANSPARENT, shine = E_MODEL_NOVA_STAR_SHINE},
    },
    --["star road"] = {
    --    ["vanilla moveset"] = {normal = E_MODEL_COSMIC_STAR, transparent = E_MODEL_COSMIC_STAR_TRANSPARENT, shine = E_MODEL_COSMIC_STAR_SHINE},
    --}
}

---@class AllNovaStars
    ---@field [string] RomhackNovaStars

---@class RomhackNovaStars
    ---@field [string] MovesetNovaStars

---@class MovesetNovaStars
    ---@field [LevelNum] LevelNovaStars

---@class LevelNovaStars
    ---@field [integer] NovaStar

---@class NovaStar
    ---@field pos Vec3f Position of the nova star
    ---@field area integer Area the ova star is in
    ---@field starID integer The star ID of the nova star, starting at 0
    ---@field challenge ChallengeType The challenge used for this nova star
    ---@field coinStar integer? Optional; The amount of coins it takes for a coin star to spawn (0 if there isn't an argument)
    ---@field warpnode integer? Optional; The warpnode to warp to, defaults to 0xA
    ---@field forceModel ModelExtendedId? Optional; Force a different model onto the uncollected nova star
    ---@field initFunc (fun(o: Object):nil)? Optional; Passes the nova star, runs on sync valid
    ---@field redsPos [Vec3f]? Challenge; Position of each red coin
    ---@field lava {startBelowY: number, speed: number}? Challenge; How far below Mario the lava starts and how fast it moves
    ---@field lampPos [Vec3f]? Challenge; Position of each lamp
    ---@field quicksand {belowY: number}? Challenge; How far below Mario the quicksand rises

---@type AllNovaStars
gNovaStarData = {
    ["vanilla"] = { -- romhack
        ["omm rebirth 1.4"] = { -- moveset
            [LEVEL_BOB] = {
                {
                    pos = {x = -1863, y = 7590, z = -752},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_CAPLESS,
                    initFunc = nil, --* this passes an object
                }
            },
            [LEVEL_WF] = {
                {
                    pos = {x = 0, y = 5000, z = 0},
                    redsPos = {
                        {x = 4408, y = 7800, z = 4500},
                        {x = -6000, y = 7400, z = 6500},
                        {x = 350, y = 2800, z = -800},
                        {x = -8000, y = 1000, z = 8000},
                        {x = -8000, y = 1000, z = -8000},
                        {x = -2965, y = 0, z = -4872},
                    },
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_CAPLESS | CHALLENGE_ABC,
                    initFunc = nil
                },
            },
            [LEVEL_WDW] = {
                {
                    pos = {x = 1126, y = -2500, z = -2201}, --* This is Squishy's collectible star that's hidden in WDW.
                    area = 2,
                    starID = 0,
                    challenge = 0,
                    forceModel = E_MODEL_SQUISHY_STAR,
                    initFunc = nil,
                }
            },
            [LEVEL_LLL] = {
                {
                    pos = {x = -3000, y = 4100, z = 4250},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_CAPLESS,
                    initFunc = nil,
                },
            },
            [LEVEL_PSS] = {
                {
                    pos = {x = -7000, y = 6700, z = 7000},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_TIMED,
                    initFunc = function (o)
                        o.oChallengeTimerMax = 500
                    end,
                }
            },
            [LEVEL_BBH] = {
                {
                    pos = {x = 3400, y = 870, z = 150},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_CAPLESS,
                    initFunc = nil,
                }
            },
            [LEVEL_THI] = {
                {
                    pos = {x = -8200, y = 6850, z = -2555},
                    area = 1,
                    starID = 0,
                    challenge = 0,
                    initFunc = nil,
                }
            },
            [LEVEL_TTM] = {
                {
                    pos = {x = 1255, y = 6000, z = 575},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_TIMED,
                    initFunc = function (o)
                        o.oChallengeTimerMax = 460
                    end,
                }
                
            },
            [LEVEL_WMOTR] = {
                {
                    pos = {x = 73, y = 2400, z = -124},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_COIN_REQUIREMENT | CHALLENGE_ACID_RAIN | CHALLENGE_CAPLESS | CHALLENGE_TIMED,
                    initFunc = function (o)
                        o.oCoinReq = 56
                        o.oChallengeTimerMax = 1200
                    end,
                }
            },
            [LEVEL_DDD] = {
                {
                    pos = {x = 1319, y = -4000, z = -3437},
                    area = 2,
                    starID = 0,
                    challenge = CHALLENGE_CAPLESS,
                    initFunc = nil,
                }
            },
            [LEVEL_COTMC] = {
                {
                    pos = {x = 0, y = 4030, z = -7420},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_GROUNDLESS | CHALLENGE_TIMED,
                    initFunc = function (o)
                        o.oChallengeTimerMax = 450
                    end,
                }
            },
            [LEVEL_TOTWC] = {
                {
                    pos = {x = 0, y = 6600, z = 0},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_CAPLESS | CHALLENGE_COIN_REQUIREMENT,
                    initFunc = function (o)
                        o.oCoinReq = 63
                    end,
                }
            },
		    [LEVEL_BITFS] = {
                {
                    pos = {x = 6750, y = 3250, z = 100},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_TIMED,
                    initFunc = function (o)
                        o.oChallengeTimerMax = 550
                    end,
                }
            },
            [LEVEL_BITS] = {
                {
                    pos = {x = 355, y = 6653, z = -5324},
                    redsPos = {
                        {x =  1590, y = -1120, z = -2055},
                        {x = -2960, y = -3360, z =     0},
                        {x =  -298, y = -2600, z = -1600},
                        {x =  7618, y =  4850, z = -1909},
                        {x =  -136, y =  3109, z =  -909},
                        {x = -3330, y =  4050, z =  -900},
                        {x = -4830, y =  2103, z = -1030},
                        {x =   348, y =  5921, z = -4585},
                    },
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_GROUNDLESS,
                    initFunc = function (o)
                        for_each_object_with_behavior(id_bhvRedCoin, obj_mark_for_deletion)
                    end,
                }
            },
            [LEVEL_SL] = {
                {
                    pos = {x = 4482, y = 307, z = 1170},
                    area = 1,
                    starID = 0,
                    coinStar = 100,
                    challenge = 0,
                    initFunc = nil,
                }
            },
            [LEVEL_RR] = {
                {
                    pos = {x = -7747, y = 4126, z = -3806},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_CAPLESS,
                    initFunc = function (o)
                        spawn_non_sync_object(id_bhvPushableMetalBox, E_MODEL_METAL_BOX, -7747, 3576, -3404, function (obj)
                            obj.oFaceAngleYaw = 0
                            obj.oMoveAngleYaw = 0
                        end)
                    end,
                }
            },
            [LEVEL_JRB] = {
                {
                    pos = {x = 1800, y = 6200, z = 1700},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_CAPLESS,
                    initFunc = nil,
                }
            },
            [LEVEL_CCM] = {
                {
                    pos = {x = -500, y = 800, z = -400},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_CAPLESS,
                    initFunc = nil,
                }
            },
            [LEVEL_HMC] = {
                {
                    pos = {x = -7590, y = 2674, z = 7557},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_COSMIC_CLONE | CHALLENGE_COIN_REQUIREMENT | CHALLENGE_TOXIC_GAS | CHALLENGE_CAPLESS,
                    initFunc = function (o)
                        o.oCoinReq = 139
                    end,
                }
            },
        },
        ["vanilla moveset"] = {
            [LEVEL_BOB] = {
                {
                    pos = {x = 3304, y = 4834, z = -4603},
                    redsPos = {
                        {x = 4408, y = 832, z = 1269},
                        {x = 4064, y = 1853, z = -2328},
                        {x = 7149, y = 2040, z = -6710},
                        {x = 4399, y = 3136, z = 829},
                        {x = -6012, y = 768, z = 2418},
                        {x = -2965, y = 0, z = -4872},
                    },
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_ACID_RAIN,
                },
                {
                    pos = {x = 1552, y = 1168, z = 302},
                    area = 1,
                    starID = 1,
                    challenge = 0,
                    initFunc = function ()   -- A L'AIDE BLOCKY i gotchu bruh
                       for_each_object_with_behavior(id_bhvChainChomp, obj_mark_for_deletion)
                       for_each_object_with_behavior(id_bhvBobomb, obj_mark_for_deletion)
                       for_each_object_with_behavior(id_bhvBreakableBoxSmall, obj_mark_for_deletion)
                    end,
                },
            },
            [LEVEL_BITDW] = {
                {
                    pos = {x = 7186, y = 2931, z = -13},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_PVP_RACE,
                    initFunc = function ()
                        for_each_object_with_behavior(id_bhvWarpPipe, obj_mark_for_deletion)
                    end
                },
            },
            [LEVEL_WF] = {
                {
                    pos = {x = 4600, y = 556, z = 1152},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_CAPLESS | CHALLENGE_COSMIC_CLONE | CHALLENGE_DAREDEVIL | CHALLENGE_COIN_REQUIREMENT,
                    initFunc = function (o)
                        o.oCoinReq = 130
                    end
                },
            },
            [LEVEL_JRB] = {
                {
                    pos = {x = 3587, y = 1816, z = 5023},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_DARKNESS | CHALLENGE_COIN_REQUIREMENT | CHALLENGE_TOXIC_GAS,
                    initFunc = function (o)
                        o.oCoinReq = 60
                        for_each_object_with_behavior(id_bhvExclamationBox, obj_mark_for_deletion)
                        for_each_object_with_behavior(id_bhvBlueCoinSwitch, obj_mark_for_deletion)
                        for_each_object_with_behavior(id_bhvClamShell, obj_mark_for_deletion)
                        set_water_level(0, -9999, false)
                        set_water_level(1, -9999, false)
                        set_water_level(2, -9999, false)
                    end
                },
            },
            [LEVEL_CCM] = {
                {
                    pos = {x = -2024, y = 2300, z = -1539},
                    area = 1,
                    starID = 0,
                    challenge = 0,
                    initFunc = function ()
                        for_each_object_with_behavior(id_bhvFadingWarp, obj_mark_for_deletion)
                    end,
                },
                {
                    pos = {x = -60, y = 5680, z = -156},
                    lava = {startBelowY = 1000, speed = 11},
                    area = 1,
                    starID = 1,
                    challenge = CHALLENGE_RISING_LAVA,
                    initFunc = function ()
                        local m = gMarioStates[0]

                        m.pos.x = 5346
                        m.pos.y = -4607
                        m.pos.z = 1022
                        soft_reset_camera(m.area.camera)
                    end,
                }
            },
            [LEVEL_BBH] = {
                {
                    pos = {x = -2706, y = -2133, z = -1474},
                    area = 1,
                    starID = 0,
                    challenge = 0,
                    initFunc = nil,
                }
            },
            [LEVEL_HMC] = {
                {
                    pos = {x = 670, y = -3975, z = 6015},
                    area = 1,
                    starID = 0,
                    challenge = 0,
                    initFunc = nil,
                }
            },
            [LEVEL_LLL] = {
                {
                    pos = {x = 1747, y = 20, z = -5682},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_ABC | CHALLENGE_DAREDEVIL,
                    initFunc = function (o)
                        o.oAPressMaxCount = 0
                    end
                },
            },
            [LEVEL_SSL] = {
                {
                    pos = {x = 507, y = 4959, z = -559},
                    lampPos = {
                        {x = 0, y = 0, z = 4407},
                        {x = -3554, y = 0, z = 3676},
                        {x = 3554, y = 0, z = 3676},
                        {x = 3554, y = 0, z = -3804},
                        {x = -3554, y = 0, z = -3804},

                        {x = 54, y = 435, z = -3505},
                        {x = 3331, y = 640, z = -936},

                        {x = 1996, y = 1280, z = 2863},
                        {x = -2503, y = 1920, z = 2863},
                        {x = -2503, y = 1920, z = -2303},

                        {x = 1732, y = 3200, z = -2288},

                        {x = -759, y = 3942, z = 1247},
                        {x = -852, y = 4045, z = -916},
                    },
                    area = 2,
                    starID = 0,
                    challenge = 0,
                    initFunc = nil,
                },
                {
                    pos = {x = 1794, y = 1148, z = -2557},
                    redsPos = {
                        {x = 4876, y = 0, z = -835},
                        {x = 6392, y = 0, z = -2273},
                        {x = 5914, y = 0, z = -4858},
                        {x = 3336, y = 0, z = -4851},
                        {x = -1264, y = 0, z = -6384},
                        {x = -5878, y = 1024, z = -2565},
                        {x = -2044, y = 1103, z = -527},
                        {x = 1794, y = 1024, z = 771},
                        {x = 5942, y = 614, z = 3281},
                        {x = -794, y = 256, z = 752},
                    },
                    area = 1,
                    starID = 1,
                    challenge = CHALLENGE_ACID_RAIN | CHALLENGE_CAPLESS,
                    initFunc = function (o)
                        for_each_object_with_behavior(id_bhvExclamationBox, obj_mark_for_deletion)
                        for_each_object_with_behavior(id_bhvStar, obj_mark_for_deletion)
                        for_each_object_with_behavior(id_bhvKlepto, obj_mark_for_deletion)
                    end
                },
            },
            [LEVEL_SL] = {
                {
                    pos = {x = 0, y = 5620, z = 0},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_DAREDEVIL,
                    initFunc = function (o)
                        gMarioStates[0].area.terrainType = TERRAIN_SLIDE
                        for_each_object_with_behavior(id_bhvCannon, obj_mark_for_deletion)
                        for_each_object_with_behavior(id_bhvCannonBarrel, obj_mark_for_deletion)
                        for_each_object_with_behavior(id_bhvExclamationBox, obj_mark_for_deletion)
                    end
                }
            },
            [LEVEL_WDW] = {
                {
                    pos = {x = -770, y = 3360, z = -3593},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_ACID_RAIN | CHALLENGE_COIN_REQUIREMENT,
                    initFunc = function (o)
                        o.oCoinReq = 100
                        spawn_sync_object(id_bhvStaticCheckeredPlatform, E_MODEL_CHECKERBOARD_PLATFORM, 3992, 2720, -3207, function (box)
                            obj_scale_xyz(box, 2.75, 4, 4)
                            box.oCollisionDistance = 3000
                        end)
                    end
                },
                {
                    pos = {x = -1957, y = -1141, z = -3316},
                    area = 1,
                    starID = 1,
                    challenge = 0,
                    initFunc = nil
                }
            },
            [LEVEL_TTM] = {
                {
                    pos = {x = 1339, y = -5444, z = -1632},
                    area = 1,
                    starID = 0,
                    challenge = 0,
                    initFunc = function (o)
                        spawn_sync_object(id_bhvStaticCheckeredPlatform, E_MODEL_CHECKERBOARD_PLATFORM, 3083, 1115, -1061, function (box)
                            obj_scale_xyz(box, 1, 9, 1)
                            box.oFaceAngleYaw = -15092
                            box.oMoveAngleYaw = -15092
                            box.oCollisionDistance = 3000
                        end)
                    end,
                },
                {
                    pos = {x = 1776, y = 1281, z = 1041},
                    redsPos = {
                        {x = -2896, y = -2733, z = 3674},
                        {x = 7352, y = -2989, z = 1442},
                        {x = -3235, y = -2447, z = -3693},
                        {x = 2521, y = -1295, z = 1363},
                        {x = 1026, y = 2545, z = 294},
                    },
                    area = 1,
                    starID = 1,
                    challenge = CHALLENGE_COSMIC_CLONE | CHALLENGE_DAREDEVIL | CHALLENGE_TIMED,
                    initFunc = function (o)
                        o.oChallengeTimerMax = 2570
                        spawn_sync_object(id_bhvStaticCheckeredPlatform, E_MODEL_CHECKERBOARD_PLATFORM, 3083, 1115, -1061, function (box)
                            obj_scale_xyz(box, 1, 9, 1)
                            box.oFaceAngleYaw = -15092
                            box.oMoveAngleYaw = -15092
                            box.oCollisionDistance = 3000
                        end)
                    end,
                },
            },
            [LEVEL_THI] = {
                {
                    pos = {x = 0, y = 4100, z = -1570},
                    lava = {startBelowY = 4000, speed = 17.5},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_RISING_LAVA,
                    initFunc = nil,
                },
                {
                    pos = {x = -5864, y = -3419, z = -5492},
                    area = 1,
                    starID = 1,
                    challenge = 0,
                    initFunc = nil,
                }
            },
            [LEVEL_TTC] = {
                {
                    pos = {x = 2200, y = 7300, z = 2210},
                    redsPos = {
                        {x = 1520, y = -5225, z = -1351}, -- Below the spawn
                        {x = 938, y = -3350, z = -1051}, -- Top of first pendulumn
                        {x = -1300, y = -3318, z = 1354}, -- Between the 3 rotating blocks
                        {x = -416, y = -2500, z = -1907}, -- Inside alcove
                        {x = 1335, y = -2088, z = -1115}, -- Between gears
                        {x = -1854, y = -1453, z = -740}, -- Beside Roll Into the Cage
                        {x = -1100, y = 750, z = -1020}, -- Pit and the Pendulumns
                        {x = 922, y = 1000, z = 1170}, -- Between exclamation boxes
                        {x = 661, y = 1951, z = 1876}, -- 1-up after the 3 pushers
                        {x = -1305, y = 2514, z = -994}, -- Inside Timed Jumps on Moving Bars
                        {x = -390, y = 4100, z = 1890}, -- The exclamation box near the 4 pushers
                        {x = 1875, y = 3300, z = 305}, -- Near the 3 treadmills
                        {x = -567, y = 3950, z = 0}, -- Below highest center platform
                        {x = -942, y = 6075, z = -844}, -- Between highest exclamation boxes
                        {x = 2352, y = 6400, z = 2347}, -- Above whomp exclamation box
                        {x = 1919, y = 7300, z = 1919}, -- Above whomp
                    },
                    quicksand = {belowY = 850},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_QUICKSAND_PIT,
                    initFunc = nil,
                },
            },
            [LEVEL_BITFS] = {
                {
                    pos = {x = 5772, y = 4681, z = 87},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_TIMED,
                    initFunc = function (o)
                        o.oChallengeTimerMax = 50 * 30
                    end,
                }
            },
            [LEVEL_BITS] = {
                {
                    pos = {x = 366, y = 6857, z = -6044},
                    lava = {startBelowY = 3000, speed = 11},
                    area = 1,
                    starID = 0,
                    challenge = CHALLENGE_RISING_LAVA | CHALLENGE_TOXIC_GAS,
                }
            },
            --[LEVEL_CASTLE] = {
            --    {
            --        pos = {x = -212, y = 2373, z = 7137},
            --        area = 2,
            --        starID = 0,
            --        challenge = CHALLENGE_NO_BLJS,
            --    }
            --},
        },
    },
    ["star road"] = {
        ["vanilla moveset"] = {
            [LEVEL_JRB] = {
                {
                    pos = {x = 5755, y = 3756, z = 4788},
                    area = 1,
                    starID = 0,
                    challenge = 0,
                    initFunc = function (o)
                        for_each_object_with_behavior(id_bhvTree, obj_mark_for_deletion)
                        for_each_object_with_behavior(id_bhvHoot, obj_mark_for_deletion)
                    end
                },
            },
        }
    }
}

gNovaPipes = {
    ["vanilla"] = {
        ["omm rebirth 1.4"] = {
            [LEVEL_CASTLE] = {
                {pipeArea = 2, level = LEVEL_WMOTR, course = COURSE_WMOTR, pos = {x = 1641, y = 2765, z = 6619}},
                {pipeArea = 3, level = LEVEL_DDD, course = COURSE_DDD, pos = {x = 5049, y = -1074, z = 1993}},
                {pipeArea = 2, level = LEVEL_TTM, course = COURSE_TTM, pos = {x = -1012, y = 1203, z = 3548}},
                {pipeArea = 3, level = LEVEL_COTMC, course = COURSE_COTMC, pos = {x = 3468, y = -1279, z = -2661}},
                {pipeArea = 1, level = LEVEL_TOTWC, course = COURSE_TOTWC, pos = {x = -475, y = 0, z = 85}},
                {pipeArea = 3, level = LEVEL_BITFS, course = COURSE_BITFS, pos = {x = 3562, y = -1074, z = 1996}},
                {pipeArea = 1, level = LEVEL_BOB, course = COURSE_BOB, pos = {x = -5000, y = 205, z = -1024}},
                {pipeArea = 1, level = LEVEL_WF, course = COURSE_WF, pos = {x = 259, y = -306, z = -3002}},
                {pipeArea = 1, level = LEVEL_PSS, course = COURSE_PSS, pos = {x = 1709, y = 614, z = -2543}},
                {pipeArea = 1, level = LEVEL_JRB, course = COURSE_JRB, pos = {x = 3654, y = 307, z = 159}},
                {pipeArea = 1, level = LEVEL_CCM, course = COURSE_CCM, pos = {x = -2295, y = -409, z = -3390}},
                {pipeArea = 3, level = LEVEL_LLL, course = COURSE_LLL, pos = {x = -969, y = -1177, z = -3574}},
                {pipeArea = 3, level = LEVEL_HMC, course = COURSE_HMC, pos = {x = 1897, y = -1381, z = -2662}},
                {pipeArea = 2, level = LEVEL_THI, course = COURSE_THI, pos = {x = -6856, y = 1306, z = 1825}},
                {pipeArea = 2, level = LEVEL_SL, course = COURSE_SL, pos = {x = 3549, y = 1408, z = 3544}},
                {pipeArea = 2, level = LEVEL_RR, course = COURSE_RR, pos = {x = -2087, y = 2765, z = 6658}},
                {pipeArea = 2, level = LEVEL_WDW, course = COURSE_WDW, pos = {x = 1852, y = 2253, z = 7094}},
                {pipeArea = 2, level = LEVEL_BITS, course = COURSE_BITS, pos = {x = -201, y = 5018, z = -1786}},
            },
            [LEVEL_CASTLE_COURTYARD] = {
                {pipeArea = 1, level = LEVEL_BBH, course = COURSE_BBH, pos = {x = 0, y = 0, z = -3562}},
            }
        },
        ["vanilla moveset"] = {
            [LEVEL_CASTLE] = {
                {pipeArea = 1, level = LEVEL_BOB, course = COURSE_BOB, pos = {x = -5000, y = 205, z = -1024}},
                {pipeArea = 1, level = LEVEL_WF, course = COURSE_WF, pos = {x = 259, y = -306, z = -3002}},
                {pipeArea = 1, level = LEVEL_JRB, course = COURSE_JRB, pos = {x = 3928, y = 307, z = 222}},
                {pipeArea = 1, level = LEVEL_CCM, course = COURSE_CCM, pos = {x = -2295, y = -409, z = -3390}},
                {pipeArea = 1, level = LEVEL_BITDW, course = COURSE_BITDW, pos = {x = -4252, y = 717, z = -3058}},
                {pipeArea = 3, level = LEVEL_HMC, course = COURSE_HMC, pos = {x = 1897, y = -1381, z = -2662}},
                {pipeArea = 3, level = LEVEL_LLL, course = COURSE_LLL, pos = {x = -1053, y = -1177, z = -3516}},
                {pipeArea = 3, level = LEVEL_SSL, course = COURSE_SSL, pos = {x = -2867, y = -1279, z = -3076}},
                {pipeArea = 2, level = LEVEL_SL, course = COURSE_SL, pos = {x = 3259, y = 1613, z = 290}},
                {pipeArea = 2, level = LEVEL_WDW, course = COURSE_WDW, pos = {x = -255, y = 1203, z = 208}},
                {pipeArea = 2, level = LEVEL_TTM, course = COURSE_TTM, pos = {x = -1012, y = 1203, z = 3548}},
                {pipeArea = 2, level = LEVEL_THI, course = COURSE_THI, pos = {x = -6619, y = 1306, z = 1828}},
                {pipeArea = 2, level = LEVEL_TTC, course = COURSE_TTC, pos = {x = -696, y = 2253, z = 7049}},
                {pipeArea = 3, level = LEVEL_BITFS, course = COURSE_BITFS, pos = {x = 3562, y = -1074, z = 1996}},
                {pipeArea = 2, level = LEVEL_BITS, course = COURSE_BITS, pos = {x = -201, y = 5018, z = -1786}},
                --{pipeArea = 2, level = LEVEL_CASTLE, course = COURSE_NONE, pos = {x = 864, y = 1203, z = 2540}},
            },
            [LEVEL_CASTLE_COURTYARD] = {
                {pipeArea = 1, level = LEVEL_BBH, course = COURSE_BBH, pos = {x = 0, y = 0, z = -3562}},
            }
        },
    },
    --["star road"] = {
    --    ["vanilla moveset"] = {
    --        [LEVEL_CASTLE_GROUNDS] = {
    --            {pipeArea = 1, level = LEVEL_JRB, course = COURSE_JRB, pos = {x = 2993, y = -614, z = 1714}},
    --        }
    --    }
    --}
}

novaStarActSelect = {
    ["vanilla"] = {
        ["omm rebirth 1.4"] = {
            [LEVEL_WMOTR] = {
                {name = "Acidic Clouds"},
            },
            [LEVEL_DDD] = {
                {name = "Behind The Caged Pipe"},
            },
            [LEVEL_TTM] = {
                {name = "Speeding Up The Mountain"},
            },
            [LEVEL_COTMC] = {
                {name = "Wallkicking Up The Cavern"},
            },
            [LEVEL_TOTWC] = {
                {name = "All Coins, No Flight"},
            },
            [LEVEL_BITFS] = {
                {name = "Make A Quick Way Under The Warp"},
            },
            [LEVEL_BITS] = {
                {name = "The Groundless 8 Green Coins"},
            },
            [LEVEL_HMC] = {
                {name = "Cosmic Clone Catastrophe"},
            },
            [LEVEL_PSS] = {
                {name = "Chop Chop You're on the Clock!"},
            },
            [LEVEL_CCM] = {
                {name = "The Mystery Inside the Mountain"},
            },
            [LEVEL_BOB] = {
                {name = "Hyper Speeding Double Jump"},
            },
            [LEVEL_WF] = {
                {name = "Jumpless Secret Coins"},
            },
            [LEVEL_LLL] = {
                {name = "Two Frame Dilemma"},
            },
            [LEVEL_BBH] = {
                {name = "Mysterious Clip into the Mansion"},
            },
            [LEVEL_RR] = {
                {name = "Crazy Bob-omb Carrying"},
            },
            [LEVEL_THI] = {
                {name = "Wall Bonk into the Unknown"},
            },
            [LEVEL_SL] = {
                {name = "Cannon With the Sparkly Star"},
            },
            [LEVEL_JRB] = {
                {name = "Pole Jumping for Height"},
            },
            [LEVEL_WDW] = {
                {name = "Squishamamishable, The Squishiest Secret Star"},
            },
        },
        ["vanilla moveset"] = {
            [LEVEL_BOB] = {
                {name = "The Acidic Coins"},
                {name = "Behind Chain Chomp's Gate?!"},
            },
            [LEVEL_WF] = {
                {name = "Devilish Collectathon"},
            },
            [LEVEL_JRB] = {
                {name = "The Dark, Dark Bay"},
            },
            [LEVEL_CCM] = {
                {name = "The One Unit Gap"},
                {name = "Top, Top Of The Mountain"},
            },
            [LEVEL_BITDW] = {
                {name = "Footrace With Blocky the.. mario?"},
            },
            [LEVEL_BBH] = {
                {name = "Behind The Stone Pillar"},
            },
            [LEVEL_HMC] = {
                {name = "The Secret Invisible Room"},
            },
            [LEVEL_LLL] = {
                {name = "No Jumps?"},
            },
            [LEVEL_SSL] = {
                {name = "Obscure Pyramid Climb", sendPlayerToArea = 2},
                {name = "Sandy Green Coins"},
            },
            [LEVEL_SL] = {
                {name = "Slip Sliding Again"}
            },
            [LEVEL_WDW] = {
                {name = "The Mystery Of Wet Or Die"},
                {name = "The Tunnel's Mystery"},
            },
            [LEVEL_TTM] = {
                {name = "Core of the Mountain"},
                {name = "The Cosmic Marathon"},
            },
            [LEVEL_THI] = {
                {name = "Rapid Eruption"},
                {name = "Hidden Below The Plants"},
            },
            [LEVEL_TTC] = {
                {name = "Don't Fall Down"},
            },
            [LEVEL_BITFS] = {
                {name = "The Clock's Ticking!"},
            },
            [LEVEL_BITS] = {
                {name = "A Race Against Fire"},
            },
            --[LEVEL_CASTLE] = {
            --    {name = "Tick Tock Clock's Secret", sendPlayerToArea = 2},
            --},
        }
    },
    --["star road"] = {
    --    ["vanilla moveset"] = {
    --        [LEVEL_JRB] = {
    --            {name = "Hootless Climb"},
    --        },
    --    }
    --}
}

gNovaStarChallengeNames = {
    [CHALLENGE_TIMED] = "Timer",
    [CHALLENGE_TOXIC_GAS] = "Toxic Gas",
    [CHALLENGE_COSMIC_CLONE] = "Cosmic Clones",
    [CHALLENGE_COIN_REQUIREMENT] = "Coin Requirement",
    [CHALLENGE_CAPLESS] = "No Powerups",
    [CHALLENGE_RISING_LAVA] = "Rising Lava",
    [CHALLENGE_DAREDEVIL] = "Daredevil",
    [CHALLENGE_ACID_RAIN] = "Acid Rain",
    [CHALLENGE_QUICKSAND_PIT] = "Quicksand Pit",
    [CHALLENGE_PVP_RACE] = "Player Race",
    [CHALLENGE_GROUNDLESS] = "Groundless",
    [CHALLENGE_DARKNESS] = "Darkness",
    [CHALLENGE_ABC] = "Limited A Presses",
    [CHALLENGE_NO_BLJS] = "Capped Negative Speed",
}
