E_MODEL_NOVA_PIPE = smlua_model_util_get_id("nova_pipe_geo")
TEXTURE_NOVA_PIPE = get_texture_info("smallPipe")

TEXTURE_DAREDEVIL = get_texture_info("dareDevil")
TEXTURE_DAREDEVIL_OVERLAY = get_texture_info("dareDevilOverlay")
TEXTURE_DAREDEVIL_DEAD = get_texture_info("dareDevil0")

E_MODEL_LAVA = smlua_model_util_get_id("lava_geo")
COL_LAVA = smlua_collision_util_get("lava_collision")

E_MODEL_QUICKSAND = smlua_model_util_get_id("quicksand_geo")
COL_QUICKSAND = smlua_collision_util_get("quicksand_collision")

E_MODEL_LAMP = smlua_model_util_get_id("lamp_geo")
E_MODEL_GREEN_COIN = smlua_model_util_get_id("green_coin_geo")

SAMPLE_ZTAR_FANFARE = audio_sample_load("ztar_fanfare.mp3") --* fanfare by 6b

ps = gPlayerSyncTable
np = gNetworkPlayers

lightR = 255
lightG = 255
lightB = 255

MAX_LIGHT_DISTANCE = 1000

preNovaStarCoinStarReq = gLevelValues.coinsRequiredForCoinStar
preNovaStarDisableActs = gLevelValues.disableActs

debug = false

define_custom_obj_fields({
    oF32Value = "f32"
})

function degrees_to_sm64(degrees)
    return math.floor(degrees * 65535 / 360)
end

function shorten_string(input)
    local shortened = ""
    for i = 1, #input, 2 do
        shortened = shortened .. input:sub(i, i)
    end
    return shortened
end

function count_string_chars(str)
    local count = 0
    for _ in string.gmatch(str, ".") do
        count = count + 1
    end
    return count
end

currHack = "vanilla"
currMoveset = "vanilla moveset"

for i, mod in pairs(gActiveMods) do
    if mod.enabled then
        if mod.incompatible and mod.incompatible:find("romhack") then
            currHack = mod.name:lower()
        end
        if mod.incompatible and mod.incompatible:find("moveset") then
            currMoveset = mod.name:lower()
        end
    end
end

modStorageKey = shorten_string(currHack) .. shorten_string(currMoveset)

CHALLENGE_TIMER_MAX = 0xFFFF
CHALLENGE_TIMER_SHIFT = 32
CHALLENGE_TIMER_MASK = 0xFFFF << CHALLENGE_TIMER_SHIFT

CHALLENGE_FLAGS_MASK = 0xFFFF     -- 32 bits for challenge flags

LEVEL_ACT_SELECT = level_register('level_actSelect_entry', COURSE_NONE, 'Act Select', 'actSelect', 28000, 0x28, 0x28, 0x28)

targetWarpInfo = {
    targetCourse = COURSE_BOB,
    targetLevel = LEVEL_BOB,
    targetArea = 1,
}

MAX_ACT_NUM = 1

for i = 0, MAX_PLAYERS - 1 do
    ps[i].challenge = 0x000000 -- AAAA is timer, BBBB is the challenge
    ps[i].novaAct = 0  --* not needed but it simplifies things
    ps[i].hasTouchedGround = false
    ps[i].APressCount = 0
end

function reset_challenges(playerIndex)
    ps[playerIndex].challenge = 0x000000
    ps[playerIndex].hasTouchedGround = false
    ps[playerIndex].APressCount = 0
end

function get_challenge_timer(playerIndex)
    return (ps[playerIndex].challenge & CHALLENGE_TIMER_MASK) >> CHALLENGE_TIMER_SHIFT
end

function increment_challenge_timer(playerIndex, increment)
    local timer = get_challenge_timer(playerIndex)
    timer = math.min(timer + increment, CHALLENGE_TIMER_MAX)
    ps[playerIndex].challenge = (ps[playerIndex].challenge & ~CHALLENGE_TIMER_MASK) | (timer << CHALLENGE_TIMER_SHIFT)
end

function set_challenge_timer(playerIndex, value)
    value = math.min(value, CHALLENGE_TIMER_MAX)
    ps[playerIndex].challenge = (ps[playerIndex].challenge & ~CHALLENGE_TIMER_MASK) | (value << CHALLENGE_TIMER_SHIFT)
end

function set_challenge(playerIndex, challenge)
    ps[playerIndex].challenge = ps[playerIndex].challenge | (challenge & CHALLENGE_FLAGS_MASK)
end

function get_challenge(playerIndex, challenge)
    return (ps[playerIndex].challenge & (challenge & CHALLENGE_FLAGS_MASK)) ~= 0
end

function remove_challenge(playerIndex, challenge)
    ps[playerIndex].challenge = ps[playerIndex].challenge & ~(challenge & CHALLENGE_FLAGS_MASK)
    if challenge & CHALLENGE_TIMED ~= 0 then
        set_challenge_timer(playerIndex, 0)
    end
end

function search_for_obj(bhv, index)
    local o = obj_get_first_with_behavior_id(bhv)
    while o ~= nil do
        if o.oOwner == index then
            return o
        end
        o = obj_get_next_with_same_behavior_id(o)
    end
    return nil
end

function for_each_object_with_behavior(behavior, func_f)
    local obj = obj_get_first_with_behavior_id(behavior)
    while obj ~= nil do
        func_f(obj)
        obj = obj_get_next_with_same_behavior_id(obj)
    end
end

function get_nova_star_challenges(hack, moveset, level, act)
    local levelData = gNovaStarData[hack] and gNovaStarData[hack][moveset] and gNovaStarData[hack][moveset][level]

    if not levelData or not levelData[act] then
        return nil
    end

    return levelData[act].challenge
end

function get_nova_star_challenge_names(hack, moveset, level, act)
    local challengeFlags = get_nova_star_challenges(hack, moveset, level, act)

    local challengeNames = {}

    if not challengeFlags or challengeFlags == 0 then
        table.insert(challengeNames, "None")
        return challengeNames
    end

    for flag, name in pairs(gNovaStarChallengeNames) do
        if (challengeFlags & flag) ~= 0 then
            table.insert(challengeNames, name)
        end
    end

    return challengeNames
end

---@param level LevelNum | integer
---@return LevelNovaStars?
function get_nova_stars_in_level(level)
    if level and gNovaStarData[currHack] and gNovaStarData[currHack][currMoveset] then
        return gNovaStarData[currHack][currMoveset][level]
    end
    return nil
end

function get_nova_star_name(level, act)
    return novaStarActSelect[currHack][currMoveset][level][act].name
end

function get_level_nova_star_count(level)
    local stars = gNovaStarData[currHack][currMoveset][level]
    if not stars then
        return 0
    end
    return #stars
end

function is_challenge_active(m, challenge)
    local level = gNetworkPlayers[m.playerIndex].currLevelNum
    local area = gNetworkPlayers[m.playerIndex].currAreaIndex
    local act = gNetworkPlayers[m.playerIndex].currActNum

    if not gNovaStarData[currHack] then
        return
    elseif not gNovaStarData[currHack][currMoveset] then
        return
    end

    local novaStarInfo = gNovaStarData[currHack][currMoveset][level]
    if novaStarInfo then
        for _, star in pairs(novaStarInfo) do
            if star.area == area and ((star.starID + 7) * (network_global_index_from_local(0) + 1)) == act and get_challenge(m.playerIndex, challenge) then
                return true
            end
        end
    end
    return false
end

local LIGHTING_CHALLENGES = {
    [CHALLENGE_RISING_LAVA] = {
        lighting = {255, 100, 100},
        vertex = {255, 100, 100},
        fog = {100, 0, 0},
        skybox = BACKGROUND_PURPLE_SKY,
        skybox_color = {100, 0, 0},
    },
    [CHALLENGE_COSMIC_CLONE] = {
        lighting = {100, 100, 255},
        vertex = {100, 100, 255},
        fog = {0, 0, 100},
        skybox = BACKGROUND_GREEN_SKY,
        skybox_color = {0, 0, 100},
    },
    [CHALLENGE_ACID_RAIN] = {
        lighting = {50, 220, 50},
        vertex = {50, 220, 50},
        fog = {0, 100, 0},
        skybox = BACKGROUND_DESERT,
        skybox_color = {0, 125, 0},
    }
}

local function reset_lighting()
    return 255, 255, 255, -1
end

function update_lighting(m)
    local r, g, b = reset_lighting()
    local vr, vg, vb = r, g, b
    local fr, fg, fb = r, g, b
    local skybox = -1
    local sky_r, sky_g, sky_b = r, g, b

    local activeChallenges = {}

    for challenge, settings in pairs(LIGHTING_CHALLENGES) do
        if is_challenge_active(m, challenge) then
            table.insert(activeChallenges, settings)
        end
    end

    if #activeChallenges == 0 then
        set_lighting_color(0, r)
        set_lighting_color(1, g)
        set_lighting_color(2, b)
        set_vertex_color(0, vr)
        set_vertex_color(1, vg)
        set_vertex_color(2, vb)
        set_fog_color(0, fr)
        set_fog_color(1, fg)
        set_fog_color(2, fb)
        set_override_skybox(skybox)
        set_skybox_color(0, sky_r)
        set_skybox_color(1, sky_g)
        set_skybox_color(2, sky_b)
        lightR = r
        lightG = g
        lightB = b
        return
    end

    if #activeChallenges == 1 then
        local settings = activeChallenges[1]
        r, g, b = table.unpack(settings.lighting)
        vr, vg, vb = table.unpack(settings.vertex)
        fr, fg, fb = table.unpack(settings.fog)
        skybox = settings.skybox
        sky_r, sky_g, sky_b = table.unpack(settings.skybox_color)
    else
        local weight = 1 / #activeChallenges

        local lr, lg, lb = 0, 0, 0
        local vrr, vrg, vrb = 0, 0, 0
        local frf, frg, frb = 0, 0, 0
        local sr, sg, sb = 0, 0, 0

        for _, settings in ipairs(activeChallenges) do
            lr = lr + settings.lighting[1] * weight
            lg = lg + settings.lighting[2] * weight
            lb = lb + settings.lighting[3] * weight

            vrr = vrr + settings.vertex[1] * weight
            vrg = vrg + settings.vertex[2] * weight
            vrb = vrb + settings.vertex[3] * weight

            frf = frf + settings.fog[1] * weight
            frg = frg + settings.fog[2] * weight
            frb = frb + settings.fog[3] * weight

            sr = sr + settings.skybox_color[1] * weight
            sg = sg + settings.skybox_color[2] * weight
            sb = sb + settings.skybox_color[3] * weight
        end

        r, g, b = math.floor(lr), math.floor(lg), math.floor(lb)
        vr, vg, vb = math.floor(vrr), math.floor(vrg), math.floor(vrb)
        fr, fg, fb = math.floor(frf), math.floor(frg), math.floor(frb)
        sky_r, sky_g, sky_b = math.floor(sr), math.floor(sg), math.floor(sb)

        skybox = activeChallenges[#activeChallenges].skybox
    end

    set_lighting_color(0, r)
    set_lighting_color(1, g)
    set_lighting_color(2, b)

    set_vertex_color(0, vr)
    set_vertex_color(1, vg)
    set_vertex_color(2, vb)

    set_fog_color(0, fr)
    set_fog_color(1, fg)
    set_fog_color(2, fb)

    set_override_skybox(skybox)
    set_skybox_color(0, sky_r)
    set_skybox_color(1, sky_g)
    set_skybox_color(2, sky_b)

    lightR = r
    lightG = g
    lightB = b
end

local sCanReadActions = {
    [ACT_IDLE] = true,
    [ACT_JUMP] = true,
    [ACT_PUNCHING] = true,
    [ACT_MOVE_PUNCHING] = true,
    [ACT_WALKING] = true,
}

local function mario_can_talk(m)
    if not m then return false end

    return sCanReadActions[m.action]
end

function check_read_sign(m, o)
    if not m or not o then
        return false
    end

    if (m.input & (INPUT_B_PRESSED | INPUT_A_PRESSED)) ~= 0 and obj_check_hitbox_overlap(m.marioObj, o) then
        local facingDYaw = (o.oMoveAngleYaw + 0x8000 - m.faceAngle.y) % 0x10000
        if facingDYaw > 0x8000 then
            facingDYaw = facingDYaw - 0x10000
        end

        if facingDYaw >= -0x4000 and facingDYaw <= 0x4000 and mario_can_talk(gMarioStates[0]) then
            local angleInRadians = o.oMoveAngleYaw * (math.pi / 32768)

            local targetX = o.oPosX + 105.0 * math.sin(angleInRadians)
            local targetZ = o.oPosZ + 105.0 * math.cos(angleInRadians)

            m.marioObj.oMarioReadingSignDYaw = facingDYaw
            m.marioObj.oMarioReadingSignDPosX = targetX - m.pos.x
            m.marioObj.oMarioReadingSignDPosZ = targetZ - m.pos.z

            m.interactObj = o
            m.usedObj = o

            m.pos.y = o.oPosY

            return true
        end
    end

    return false
end