--* i'll likely rework this in the future so that nova stars use one long bit string

local tonumber, mod_storage_load, tostring, mod_storage_save, get_current_save_file_num,
      cur_obj_become_tangible, obj_set_model_extended, obj_mark_for_deletion, spawn_non_sync_object,
      hook_event, hook_behavior =
      tonumber, mod_storage_load, tostring, mod_storage_save, get_current_save_file_num,
      cur_obj_become_tangible, obj_set_model_extended, obj_mark_for_deletion, spawn_non_sync_object,
      hook_event, hook_behavior

MAX_STARS_PER_LEVEL = 7
local PACKET_SEND_STAR = 0

gGlobalSyncTable.courseData = {}

---@class Object
    ---@field oChallengeTimerMax integer
    ---@field oCoinReq integer
    ---@field oModel ModelExtendedId
    ---@field oAPressMaxCount integer
define_custom_obj_fields({
    oAPressMaxCount = "u32",
    oChallengeTimerMax = "u32",
    oCoinReq = "u32",
    oModel = "u32",
})

local function is_current_area_sync_valid() --* from visible lakitu
    --* check all connected players to see if their area sync is valid
    for i = 0, (MAX_PLAYERS - 1) do
        local np = gNetworkPlayers[i]
        if np ~= nil and np.connected and (not np.currLevelSyncValid or not np.currAreaSyncValid) then
            return false
        end
    end
    return true
end

local function obj_mark_for_deletion_on_sync(o)
    if gNetworkPlayers[0].currAreaSyncValid then obj_mark_for_deletion(o) end
end


function save_file_delete_ms_star_save(fileIndex)
    if fileIndex < 0 then
        return
    end

    for courseIndex = COURSE_NONE, COURSE_MAX - 1 do
        local key = modStorageKey.."_sFlags_" .. fileIndex .. "_" .. courseIndex

        if network_is_server() then
            mod_storage_save(key, tostring(0))
        else
            if gGlobalSyncTable.courseData[fileIndex] then
                gGlobalSyncTable.courseData[fileIndex][courseIndex] = 0
            end
        end
    end
end

local function save_file_set_ms_star_flags(fileIndex, courseIndex, starId, starData)
    starId = tonumber(starId)
    if fileIndex < 0 or courseIndex < -1 or starId < 0 then
        return
    end

    if network_is_server() then
        local key = modStorageKey.."_sFlags_" .. fileIndex .. "_" .. courseIndex
        local currentFlags = tonumber(mod_storage_load(key)) or 0
        currentFlags = currentFlags | (1 << starId)
        mod_storage_save(key, tostring(currentFlags))
    else
        if not gGlobalSyncTable.courseData[fileIndex] then
            gGlobalSyncTable.courseData[fileIndex] = {}
        end
        if not gGlobalSyncTable.courseData[fileIndex][courseIndex] then
            gGlobalSyncTable.courseData[fileIndex][courseIndex] = 0
        end
        gGlobalSyncTable.courseData[fileIndex][courseIndex] = gGlobalSyncTable.courseData[fileIndex][courseIndex] | (1 << starId)
    end
end

function save_file_get_ms_star_flags(fileIndex, courseIndex, starData)
    if fileIndex < 0 or courseIndex < -1 then
        return 0
    end

    if network_is_server() then
        --* Host : Retrieve star flags from mod storage
        local key = modStorageKey.."_sFlags_" .. fileIndex .. "_" .. courseIndex
        local savedFlags = tonumber(mod_storage_load(key)) or 0
        return savedFlags
    else
        --* Retrieve star flags from starData table
        local starFlags = gGlobalSyncTable.courseData[fileIndex] and gGlobalSyncTable.courseData[fileIndex][courseIndex] or 0
        return starFlags
    end
end

local function save_file_get_course_ms_star_count(fileIndex, courseIndex)
    if (not fileIndex or fileIndex < 0) then return 0 end

    local count = 0
    local flag = 1
    local starFlags = save_file_get_ms_star_flags(fileIndex, courseIndex)

    for i = 0, MAX_STARS_PER_LEVEL - 1 do
        if starFlags & flag ~= 0 then
            count = count + 1
        end
        flag = flag << 1
    end
    return count
end

--* Function to get the total number of collected stars in a file, works like in vanilla
local function save_file_get_total_ms_star_count(fileIndex, minCourse, maxCourse)
    if fileIndex < 0 then
        return 0
    end

    local count = 0

    if minCourse < -1 then
        minCourse = -1
    end
    if maxCourse >= COURSE_MAX then
        maxCourse = COURSE_MAX - 1
    end

    for courseIndex = minCourse, maxCourse do
        count = count + save_file_get_course_ms_star_count(fileIndex, courseIndex)
    end

    return save_file_get_course_ms_star_count(fileIndex, -1) + count
end

gGlobalSyncTable.storageStarCount = save_file_get_total_ms_star_count(get_current_save_file_num() - 1, COURSE_NONE, COURSE_MAX - 1)

--* table containing all the collected stars
--* this is for non host players as only the host will
--* use mod storage.
local function network_send_include_self(reliable, data)
    network_send(reliable, data)
    if data.id == PACKET_SEND_STAR then
        save_file_set_ms_star_flags(data.fileIndex, data.courseIndex, data.starId)
        gGlobalSyncTable.storageStarCount = save_file_get_total_ms_star_count(get_current_save_file_num() - 1, COURSE_NONE, COURSE_MAX - 1)
    end
end

--* Function to check if a star is collected
function is_ms_star_collected(fileIndex, courseIndex, starId)
    if fileIndex < 0 or courseIndex < -1 or starId < 0 then
        return false
    end

    local currentLevelStarFlags = save_file_get_ms_star_flags(fileIndex, courseIndex)

    return (currentLevelStarFlags & (1 << starId)) ~= 0
end

--* Function to collect a star or key and update mod storage
local function save_file_collect_ms_star_or_key(starId)
    local fileIndex = get_current_save_file_num() - 1
    local courseIndex = gNetworkPlayers[0].currCourseNum - 1

    if fileIndex < 0 or courseIndex < -1 then
        return
    end

    if not is_ms_star_collected(fileIndex, courseIndex, starId) then
        if gNetworkPlayers[0].currLevelNum == LEVEL_BOWSER_1 then
            --* this is where keys are handled, meaning you can also
            --* have mod storage keys!
        elseif gNetworkPlayers[0].currLevelNum == LEVEL_BOWSER_2 then

        elseif gNetworkPlayers[0].currLevelNum == LEVEL_BOWSER_3 then

        else
            network_send_include_self(true, {id = PACKET_SEND_STAR, fileIndex = fileIndex, courseIndex = courseIndex, starId = starId})
            --save_file_set_ms_star_flags(fileIndex, courseIndex, starId)
        end
    end
end

local function interact_ms_star_or_key(m, o)
    local gCurrSaveFileNum = get_current_save_file_num() - 1
    local gCurrLevelNum = gNetworkPlayers[0].currLevelNum
    if not m or not o then return false end
    -- only allow for local player
    if m.playerIndex ~= 0 then return false end

    local starIndex
    local starGrabAction = ACT_STAR_DANCE_EXIT
    local noExit = (o.oInteractionSubtype & INT_SUBTYPE_NO_EXIT) ~= 0
    local grandStar = (o.oInteractionSubtype & INT_SUBTYPE_GRAND_STAR) ~= 0

    local stayInLevelCommon = not (gCurrLevelNum == LEVEL_BOWSER_1 or gCurrLevelNum == LEVEL_BOWSER_2 or gCurrLevelNum == LEVEL_BOWSER_3)
    if stayInLevelCommon and gServerSettings.stayInLevelAfterStar then noExit = true end
    gLastCollectedStarOrKey = o.behavior == get_behavior_from_id(id_bhvBowserKey)

    if m.health >= 0x100 then
        if gServerSettings.stayInLevelAfterStar ~= 2 then
            mario_stop_riding_and_holding(m)
        end

        queue_rumble_data_mario(m, 5, 80)

        if not noExit then
            m.hurtCounter = 0
            m.healCounter = 0
            if m.capTimer > 1 then
                m.capTimer = 1
            end
        end

        if gLevelValues.starHeal then
            m.healCounter = 31
            m.hurtCounter = 0
        end

        if noExit then
            starGrabAction = ACT_STAR_DANCE_NO_EXIT
        end

        if m.action & ACT_FLAG_SWIMMING ~= 0 then
            starGrabAction = ACT_STAR_DANCE_WATER
        end

        if m.action & ACT_FLAG_METAL_WATER ~= 0 then
            starGrabAction = ACT_STAR_DANCE_WATER
        end

        if m.action & ACT_FLAG_AIR ~= 0 then
            starGrabAction = ACT_FALL_AFTER_STAR_GRAB
            if gLevelValues.floatingStarDance and m.pos.y - m.floorHeight > 1000 then
                starGrabAction = ACT_STAR_DANCE_WATER
            end
        end

        for i = 0, MAX_PLAYERS - 1 do
            local marioState = gMarioStates[i]
            if is_player_active(marioState) and marioState.marioObj ~= nil then
                spawn_non_sync_object(id_bhvStarKeyCollectionPuffSpawner, E_MODEL_NONE, marioState.marioObj.oPosX, marioState.marioObj.oPosY, marioState.marioObj.oPosZ, nil)
            end
        end

        o.oInteractStatus = INT_STATUS_INTERACTED
        m.interactObj = o
        m.usedObj = o

        starIndex = o.oBehParams >> 24

        save_file_collect_ms_star_or_key(starIndex & 0xFF)

        if not noExit then
            drop_queued_background_music()
            fadeout_level_music(126)
        end

        play_sound(SOUND_MENU_STAR_SOUND, m.marioObj.header.gfx.cameraToObject)
        update_mario_sound_and_camera(m)

        if grandStar then
            return set_mario_action(m, ACT_JUMBO_STAR_CUTSCENE, 0)
        end

        if noExit and gServerSettings.stayInLevelAfterStar == 2 then
            return true
        end

        if _G.OmmApi then
            starGrabAction = _G.OmmApi["ACT_OMM_STAR_DANCE"]
        end

        return set_mario_action(nearest_mario_state_to_object(o), starGrabAction, (noExit and 1 or 0) + (grandStar and 2 or 0))
    end

    network_send_object(o, true)

    return false
end

INTERACT_MS_STAR_OR_KEY = 0

local interactionTable = {
    [INTERACT_MS_STAR_OR_KEY] = interact_ms_star_or_key
}

local sCollectMSStarHitbox = {
    interactType      = INTERACT_WATER_RING,
    downOffset        = 0,
    damageOrCoinValue = 0,
    health            = 0,
    numLootCoins      = 0,
    radius            = 80,
    height            = 50,
    hurtboxRadius     = 0,
    hurtboxHeight     = 0,
}

--- @param obj Object
local function obj_set_hitbox2(obj, hitbox) --* the exposed function doesn't work for some reason??
    if obj == nil or hitbox == nil then return end
    if (obj.oFlags & OBJ_FLAG_30) == 0 then
        obj.oFlags = obj.oFlags | OBJ_FLAG_30

        obj.oInteractType = hitbox.interactType
        obj.oDamageOrCoinValue = hitbox.damageOrCoinValue
        obj.oHealth = hitbox.health
        obj.oNumLootCoins = hitbox.numLootCoins

        cur_obj_become_tangible()
    end

    obj.hitboxRadius = obj.header.gfx.scale.x * hitbox.radius
    obj.hitboxHeight = obj.header.gfx.scale.y * hitbox.height
    obj.hurtboxRadius = obj.header.gfx.scale.x * hitbox.hurtboxRadius
    obj.hurtboxHeight = obj.header.gfx.scale.y * hitbox.hurtboxHeight
    obj.hitboxDownOffset = obj.header.gfx.scale.y * hitbox.downOffset
end

local ogTempo = sequence_player_get_tempo(SEQ_PLAYER_LEVEL)

local function bhv_collect_ms_star_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    local starId = (o.oBehParams >> 24) & 0xFF
    local fileIndex = get_current_save_file_num() - 1
    local courseIndex = gNetworkPlayers[0].currCourseNum - 1

    local customModelInfo = gNovaStarModels[currHack][currMoveset]
    ogTempo = sequence_player_get_tempo(SEQ_PLAYER_LEVEL)

    if is_ms_star_collected(fileIndex, courseIndex, starId) then
        obj_set_model_extended(o, customModelInfo.transparent)
    else
        obj_set_model_extended(o, o.oModel)
        if customModelInfo.shine then
            for i = 0, 3 do
                spawn_non_sync_object(id_bhvStarShine, customModelInfo.shine, 0, 0, 0, function (shine)
                    shine.parentObj = o
                    shine.oFaceAngleRoll = math.random(-32767, 32767) * i
                    shine.oF32Value = math.random()
                    shine.oBehParams2ndByte = i
                end)
            end
        end
    end
    obj_set_hitbox2(o, sCollectMSStarHitbox)
    network_init_object(o, true, {"oBehParams"})
end

local function star_handle_challenges(o, m)
    local customModelInfo = gNovaStarModels[currHack][currMoveset]

    local model = o.oModel
    local intangible = false
    local playSound = false
    local starId = (o.oBehParams >> 24) & 0xFF
    local fileIndex = get_current_save_file_num() - 1
    local courseIndex = gNetworkPlayers[0].currCourseNum - 1

    if o.oCoinReq ~= 0 and is_challenge_active(m, CHALLENGE_COIN_REQUIREMENT) then
        if m.numCoins < o.oCoinReq then
            model = customModelInfo.transparent
            intangible = true
        else
            playSound = true
            remove_challenge(0, CHALLENGE_COIN_REQUIREMENT)
        end
    end
    
    if o.oCoinReq == 0 or m.numCoins >= o.oCoinReq then
        if is_ms_star_collected(fileIndex, courseIndex, starId) then
            model = customModelInfo.transparent
        end
    end
    

    obj_set_model_extended(o, model)
    if intangible then
        cur_obj_become_intangible()
    else
        obj_become_tangible(o)
    end

    if playSound then
        play_sound(SOUND_GENERAL_GRAND_STAR_JUMP, gGlobalSoundSource)
    end

    if is_challenge_active(m, CHALLENGE_TIMED) then
        local timeLeft = o.oChallengeTimerMax - get_challenge_timer(m.playerIndex)

        if timeLeft <= 300 and timeLeft > 0 then
            local currentTempo = sequence_player_get_tempo(SEQ_PLAYER_LEVEL)
            local currentLvlTrans = sequence_player_get_transposition(SEQ_PLAYER_LEVEL)
            local currentSfxTrans = sequence_player_get_transposition(SEQ_PLAYER_SFX)
            local newTempo = currentTempo + 24 -- increase by half a beat (48/2)

            if timeLeft % 90 == 0 then
                currentSfxTrans = currentSfxTrans + 1
            end

            if timeLeft % 30 == 0 then
                play_sound(SOUND_GENERAL_PENDULUM_SWING, gGlobalSoundSource)
                currentLvlTrans = currentLvlTrans + 1
            end

            sequence_player_set_tempo(SEQ_PLAYER_LEVEL, newTempo)
            sequence_player_set_transposition(SEQ_PLAYER_LEVEL, currentLvlTrans)
            sequence_player_set_transposition(SEQ_PLAYER_SFX, currentSfxTrans)
        end
        if timeLeft <= 0 and o.oChallengeTimerMax > 0 then
            reset_challenges(0)
            hud_show()
            obj_mark_for_deletion(o)
            spawn_mist_particles()
            for_each_object_with_behavior(id_bhvMSRedCoinStarSpawner, obj_mark_for_deletion)
            cur_obj_play_sound_1(SOUND_MENU_CAMERA_BUZZ)
            stop_background_music(get_current_background_music())
            sequence_player_set_transposition(SEQ_PLAYER_SFX, 0)
            hud_set_value(HUD_DISPLAY_FLAGS, hud_get_value(HUD_DISPLAY_FLAGS) & ~HUD_DISPLAY_FLAG_TIMER)
            hud_set_value(HUD_DISPLAY_TIMER, gLevelValues.pssSlideStarTime + 100)
        end
    end

    if is_challenge_active(m, CHALLENGE_PVP_RACE) then
        local racer = obj_get_first_with_behavior_id(id_bhvRacer)
        if racer and racer.oAction >= 3 then
            cur_obj_become_intangible()
        end
    end

    if is_challenge_active(m, CHALLENGE_ABC) then
        if ps[m.playerIndex].APressCount > o.oAPressMaxCount then
            obj_mark_for_deletion(o)
            spawn_mist_particles()
            cur_obj_play_sound_1(SOUND_MENU_CAMERA_BUZZ)
            reset_challenges(0)
        end
    end
end

local function bhv_collect_ms_star_loop(o)
    local m = nearest_mario_state_to_object(o)

    local starId = (o.oBehParams >> 24) & 0xFF
    local fileIndex = get_current_save_file_num() - 1
    local courseIndex = gNetworkPlayers[0].currCourseNum - 1

    local customModelInfo = gNovaStarModels[currHack][currMoveset]

    if is_ms_star_collected(fileIndex, courseIndex, starId) then
        model =customModelInfo.transparent
    end

    star_handle_challenges(o, m)

    if o.oIntangibleTimer >= 0 or is_challenge_active(m, CHALLENGE_PVP_RACE) then --* still have the star spin when it becomes intangible on cosmic races
        o.oFaceAngleYaw = o.oFaceAngleYaw + 0x800
    end

    if (o.oInteractStatus & INT_STATUS_INTERACTED) ~= 0 then
        interactionTable[INTERACT_MS_STAR_OR_KEY](m, o)
        obj_mark_for_deletion(o)
        o.oInteractStatus = 0
        reset_challenges(0)
        sequence_player_set_transposition(SEQ_PLAYER_LEVEL, 0)
        sequence_player_set_transposition(SEQ_PLAYER_SFX, 0)
        sequence_player_set_tempo(SEQ_PLAYER_LEVEL, ogTempo)
        hud_show()
    end
end

local function set_home_to_mario(o)
    local parentIsMario = false
    for i = 0, MAX_PLAYERS - 1 do
        if o.parentObj == gMarioStates[i].marioObj then
            parentIsMario = true
            break
        end
    end
    if parentIsMario then
        o.oHomeX = o.parentObj.oPosX
        o.oHomeZ = o.parentObj.oPosZ
        o.oHomeY = o.parentObj.oPosY
    else
        local player = nearest_player_to_object(o)
        if player then
            o.oHomeX = player.oPosX
            o.oHomeZ = player.oPosZ
            o.oHomeY = player.oPosY
        end
    end
    o.oHomeY = o.oHomeY + 250.0
    o.oPosY = o.oHomeY
    local sp1C = o.oHomeX - o.oPosX
    local sp18 = o.oHomeZ - o.oPosZ
    o.oForwardVel = math.sqrt(sp1C * sp1C + sp18 * sp18) / 23.0
end

local function bhv_spawned_ms_star_loop(o)
    if o.oAction == 0 then
        local playExclamationBoxCutscene = (is_nearest_mario_state_to_object(gMarioStates[0], o) and o.oStarSpawnExtCutsceneFlags)
        local playGenericSpawnCutscene = (o.parentObj ~= nil and o.parentObj == gMarioStates[0].marioObj)
        local playCutscene = (playExclamationBoxCutscene or playGenericSpawnCutscene)

        if o.oTimer <= 1 then --* checking for 0 doesn't work for some reason
            if playCutscene and ((gMarioStates[0].action & ACT_GROUP_MASK) ~= ACT_GROUP_CUTSCENE) then
                cutscene_object(CUTSCENE_RED_COIN_STAR_SPAWN, o)
                gMarioStates[0].freeze = 60
                set_time_stop_flags(TIME_STOP_ENABLED | TIME_STOP_MARIO_AND_DOORS)
                o.activeFlags = o.activeFlags | ACTIVE_FLAG_INITIATED_TIME_STOP
            end
            o.oAngleVelYaw = 0x800
            if o.oBehParams2ndByte == 0 then
                set_home_to_mario(o)
            else
                o.oForwardVel = 0
                o.oHomeY = o.oPosY
            end
            o.oMoveAngleYaw = cur_obj_angle_to_home()
            o.oVelY = 50.0
            o.oGravity = -4.0
            spawn_mist_particles()
            network_send_object(o, true)
        end
        cur_obj_play_sound_1(SOUND_ENV_STAR)

        spawn_non_sync_object(id_bhvSparkleSpawn, 0, o.oPosX, o.oPosY, o.oPosZ, nil)

        if o.oVelY < 0 and o.oPosY < o.oHomeY then
            o.oAction = o.oAction + 1
            o.oForwardVel = 0
            o.oVelY = 20.0
            o.oGravity = -1.0
            play_power_star_jingle(true)
            network_send_object(o, true)
        end
    elseif o.oAction == 1 then
        if o.oVelY < -4.0 then
            o.oVelY = -4.0
        end
        if o.oVelY < 0 and o.oPosY < o.oHomeY then
            gObjCutsceneDone = true
            o.oVelY = 0
            o.oGravity = 0
            o.oAction = o.oAction + 1
            o.oTimer = 0
            network_send_object(o, true)
        end
        spawn_non_sync_object(id_bhvSparkleSpawn, 0, o.oPosX, o.oPosY, o.oPosZ, nil)
    elseif o.oAction == 2 then
        if o.oTimer >= 35 then
            disable_time_stop_including_mario()
            gMarioStates[0].freeze = 0
            soft_reset_camera(gMarioStates[0].area.camera)
            o.activeFlags = o.activeFlags & ~ACTIVE_FLAG_INITIATED_TIME_STOP
            o.oAction = o.oAction + 1
            network_send_object(o, true)
        end
    else
        if o.oTimer <= 1 then
            o.oStarSpawnExtCutsceneFlags = 0
        end
        obj_set_hitbox2(o, sCollectMSStarHitbox)

        local starId = (o.oBehParams >> 24) & 0xFF
        local fileIndex = get_current_save_file_num() - 1
        local courseIndex = gNetworkPlayers[0].currCourseNum - 1

        local customModelInfo = gNovaStarModels[currHack][currMoveset]

        if is_ms_star_collected(fileIndex, courseIndex, starId) then
            model = customModelInfo.transparent
        end
    
        star_handle_challenges(o, gMarioStates[0])

        if (o.oInteractStatus & INT_STATUS_INTERACTED) ~= 0 then
            interactionTable[INTERACT_MS_STAR_OR_KEY](gMarioStates[0], o)
            obj_mark_for_deletion(o)
            o.oInteractStatus = 0
        end
        if (o.oAngleVelYaw > 0x400) and o.oIntangibleTimer >= 0 then
            o.oAngleVelYaw = o.oAngleVelYaw - 0x40
        end
    end
    cur_obj_move_using_fvel_and_gravity()
    o.oFaceAngleYaw = o.oFaceAngleYaw + o.oAngleVelYaw
    o.oInteractStatus = 0
    spawn_star_number()
end

local function mario_update(m)
    m.numStars = save_file_get_total_star_count(get_current_save_file_num() -1, COURSE_NONE, COURSE_MAX - 1) + gGlobalSyncTable.storageStarCount
end

local function on_packet_receive(data)
    if data.id == PACKET_SEND_STAR then
        save_file_set_ms_star_flags(data.fileIndex, data.courseIndex, data.starId)
        gGlobalSyncTable.storageStarCount = save_file_get_total_ms_star_count(get_current_save_file_num() - 1, COURSE_NONE, COURSE_MAX - 1)
    end
end

local function sync_mod_storage_to_sync_table()
    if not network_is_server() then
        return
    end

    for fileIndex = 0, 3 do
        for courseIndex = -1, COURSE_MAX - 1 do
            local key = modStorageKey.."_sFlags_" .. fileIndex .. "_" .. courseIndex
            local savedFlags = tonumber(mod_storage_load(key)) or 0
            gGlobalSyncTable.courseData[fileIndex] = gGlobalSyncTable.courseData[fileIndex] or {}
            gGlobalSyncTable.courseData[fileIndex][courseIndex] = savedFlags
        end
    end
end

local function bhv_custom_red_coin_init(o)
    local spawner = obj_get_first_with_behavior_id(id_bhvMSRedCoinStarSpawner)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    local hitbox = get_temp_object_hitbox()
    hitbox.interactType = INTERACT_COIN
    hitbox.damageOrCoinValue = 3
    hitbox.radius = 80
    hitbox.height = 80

    obj_set_hitbox(o, hitbox)
    if spawner then
        o.parentObj = spawner
    end
end

local function bhv_custom_red_coin_loop(o)
    obj_set_billboard(o)
    local m = nearest_mario_state_to_object(o)
    o.oAnimState = o.oAnimState + 1
    if o.oInteractStatus == INT_STATUS_INTERACTED then
        obj_mark_for_deletion_on_sync(o)
        play_sound(SOUND_MENU_COLLECT_RED_COIN + ((count_objects_with_behavior(o.behavior) - 1) << 16), m.marioObj.header.gfx.cameraToObject)
        spawn_non_sync_object(id_bhvGoldenCoinSparkles, E_MODEL_SPARKLES, o.oPosX, o.oPosY, o.oPosZ, nil)
        spawn_orange_number(count_objects_with_behavior(o.behavior) - 1, 0, 50, 0)
    end
    if is_current_area_sync_valid() then
        network_send_object(o, true)
    end

    o.oInteractStatus = 0
end

local function bhv_ms_red_coin_star_spawner_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    obj_scale(o, 1.25)
    o.header.gfx.scale.z = 0.5
    o.oFaceAnglePitch = 0x4000
    o.oHomeY = o.oPosY
    o.oAction = 0
    o.oPosY = find_floor_height(o.oPosX, o.oPosY, o.oPosZ) + 50
    network_init_object(o, true, {"oAction", "oTimer"})
end

local function bhv_ms_red_coin_star_spawner_loop(o)
    o.oFaceAngleYaw = o.oFaceAngleYaw + 0x100
    if o.oAction <= 0 then
        if count_objects_with_behavior(get_behavior_from_id(id_bhvMSRedCoin)) <= 0 then
            o.oAction = 1
            spawn_sync_object(id_bhvSpawnedMSStar, E_MODEL_STAR, o.oPosX, o.oHomeY, o.oPosZ, function (star)
                star.oModel = o.oModel
                star.oBehParams = o.oBehParams
                star.oBehParams2ndByte = 2
            end)
        end
    end

    star_handle_challenges(o, gMarioStates[0])

    local customModelInfo = gNovaStarModels[currHack][currMoveset]

    obj_set_model_extended(o, customModelInfo.transparent)

    if is_current_area_sync_valid() then
        network_send_object(o, true)
    end
end

local function on_seq_load(player, seq)
    if seq == SEQ_EVENT_CUTSCENE_COLLECT_STAR and gNetworkPlayers[0].currActNum > 6 then
        audio_sample_play(SAMPLE_ZTAR_FANFARE, gMarioStates[0].pos, 0.75)
        return SEQ_COUNT
    end
end

hook_event(HOOK_ON_SEQ_LOAD, on_seq_load)
hook_event(HOOK_ON_PLAYER_CONNECTED, sync_mod_storage_to_sync_table)
hook_event(HOOK_ON_PACKET_RECEIVE, on_packet_receive)
hook_event(HOOK_MARIO_UPDATE, mario_update)

id_bhvMSRedCoin = hook_behavior(nil, OBJ_LIST_GENACTOR, true, bhv_custom_red_coin_init, bhv_custom_red_coin_loop, "bhvModStorageRedCoin")
id_bhvMSRedCoinStarSpawner = hook_behavior(nil, OBJ_LIST_GENACTOR, true, bhv_ms_red_coin_star_spawner_init, bhv_ms_red_coin_star_spawner_loop, "bhvModStorageRedCoinStarSpawner")
id_bhvMSStar        = hook_behavior(nil, OBJ_LIST_LEVEL, true, bhv_collect_ms_star_init, bhv_collect_ms_star_loop, "bhvModStorageStar")
id_bhvSpawnedMSStar = hook_behavior(nil, OBJ_LIST_LEVEL, true, bhv_collect_ms_star_init, bhv_spawned_ms_star_loop, "bhvSpawnedModStorageStar")