
local COIN_STAR_ID = -1
local SPACING = 163
local STAR_POS_Y = 290
local STAR_POS_Z = -300

local sequenceToSet = nil

local function star_select_handler_init(o)
    hud_hide()
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    o.oBehParams2ndByte = 0

    local is100coinStarCollected = save_file_get_ms_star_flags(get_current_save_file_num() - 1, targetWarpInfo.targetCourse - 1) & (1 << COIN_STAR_ID) ~= 0

    if is100coinStarCollected then
        spawn_non_sync_object(id_bhvStaticObject, E_MODEL_STAR, 377, (o.oPosY + STAR_POS_Y) - 220, o.oPosZ + STAR_POS_Z, function (coinStar)
            obj_scale(coinStar, 0.8)
        end)
    end

    local baseX = o.oPosX - ((MAX_ACT_NUM - 1) * SPACING / 2)

    for i = 0, MAX_ACT_NUM - 1 do
        local starX = baseX + (SPACING * i)
        spawn_non_sync_object(id_bhvStarSelectorStar, E_MODEL_STAR, starX, o.oPosY + STAR_POS_Y, o.oPosZ + STAR_POS_Z, function(star)
            star.parentObj = o
            star.oBehParams2ndByte = i
        end)
    end

    cur_obj_set_home_once()
end

local transitionTimer = 0
local transitionTimerMax = 20

local menuHoldKeyIndex = 0
local menuHoldKeyTimer = 0

local MENU_SCROLL_HORIZONTAL = 0
local MENU_SCROLL_VERTICAL = 1

local function handle_menu_scrolling(m, scrollDirection, o, minIndex, maxIndex)
    local stickX = m.controller.stickX
    local stickY = m.controller.stickY
    local right = m.controller.buttonDown & R_JPAD ~= 0
    local left = m.controller.buttonDown & L_JPAD ~= 0
    local up = m.controller.buttonDown & U_JPAD ~= 0
    local down = m.controller.buttonDown & D_JPAD ~= 0
    local index = 0

    if scrollDirection == MENU_SCROLL_VERTICAL then
        if stickY > 60 or up then
            index = index + 1
        elseif stickY < -60 or down then
            index = index + 2
        end
    elseif scrollDirection == MENU_SCROLL_HORIZONTAL then
        if stickX > 60 or right then
            index = index + 2
        elseif stickX < -60 or left then
            index = index + 1
        end
    end

    if ((index ~ menuHoldKeyIndex) & index) == 2 then
        if o.oBehParams2ndByte < maxIndex then
            o.oBehParams2ndByte = o.oBehParams2ndByte + 1
            play_sound(SOUND_MENU_CHANGE_SELECT, gGlobalSoundSource)
        end
    elseif ((index ~ menuHoldKeyIndex) & index) == 1 then
        if o.oBehParams2ndByte > minIndex then
            o.oBehParams2ndByte = o.oBehParams2ndByte - 1
            play_sound(SOUND_MENU_CHANGE_SELECT, gGlobalSoundSource)
        end
    end

    if menuHoldKeyTimer == 10 then
        menuHoldKeyTimer = 8
        menuHoldKeyIndex = 0
    else
        menuHoldKeyTimer = menuHoldKeyTimer + 1
        menuHoldKeyIndex = index
    end

    if (index & 3) == 0 then
        menuHoldKeyTimer = 0
    end
end

local function star_select_handler_loop(o)

    if transitionTimer > 0 then
        transitionTimer = transitionTimer + 1

        if transitionTimer >= transitionTimerMax + 5 then
            transitionTimer = 0
            camera_unfreeze()
            hud_show()
            local warpnode = 0xA
            local nova_stars = get_nova_stars_in_level(targetWarpInfo.targetLevel)
            if nova_stars then
                local star_id = o.oBehParams2ndByte
                ---@type NovaStar
                local star = nova_stars[star_id]
                if star and star.warpnode then
                    warpnode = star.warpnode
                end
            end
            warp_to_warpnode(targetWarpInfo.targetLevel, targetWarpInfo.targetArea, (o.oBehParams2ndByte + 7) * (network_global_index_from_local(0) + 1), warpnode)
            ps[0].novaAct = o.oBehParams2ndByte + 1
        end
    end

    local transColor = {r = 255, g = 255, b = 255}

    ---@type MarioState
    local m = gMarioStates[0]

    m.freeze = 1

    local minIndex = 0
    local maxIndex = MAX_ACT_NUM - 1

    targetWarpInfo.targetArea = novaStarActSelect[currHack][currMoveset][targetWarpInfo.targetLevel][o.oBehParams2ndByte + 1].sendPlayerToArea or 1

    if transitionTimer <= 0 then
        handle_menu_scrolling(m, MENU_SCROLL_HORIZONTAL, o, minIndex, maxIndex)

        if m.controller.buttonPressed & A_BUTTON ~= 0 then
            play_sound(SOUND_MENU_STAR_SOUND, gGlobalSoundSource)
            play_character_sound(m, CHAR_SOUND_LETS_A_GO)
            play_transition(WARP_TRANSITION_FADE_INTO_COLOR, transitionTimerMax, transColor.r, transColor.g, transColor.b)
            transitionTimer = 1
        end
    end
end

local function star_selector_star_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_MOVE_XZ_USING_FVEL

    local parent = o.parentObj
    local selectedAct = parent.oBehParams2ndByte

    local customModelInfo = gNovaStarModels[currHack][currMoveset]

    if save_file_get_ms_star_flags(get_current_save_file_num() - 1, targetWarpInfo.targetCourse - 1) & (1 << (o.oBehParams2ndByte)) ~= 0 then
        obj_set_model_extended(o, customModelInfo.normal)
        if customModelInfo.shine then
            for i = 0, 3 do
                spawn_non_sync_object(id_bhvStarShine, customModelInfo.shine, 0, 0, 0, function (shine)
                    shine.parentObj = o
                    shine.oFaceAngleRoll = math.random(-32767, 32767) * i
                    shine.oF32Value = math.random() * 0.25
                    shine.oBehParams2ndByte = i
                end)
            end
        end
    else
        obj_set_model_extended(o, customModelInfo.transparent)
    end
end

local function star_selector_star_loop(o)
    local parent = o.parentObj
    local selectedAct = parent.oBehParams2ndByte

    if o.oBehParams2ndByte == selectedAct then
        o.oFaceAngleYaw = o.oFaceAngleYaw + (0x800 + (transitionTimer * 500))
        cur_obj_scale(1.3)
    else
        o.oFaceAngleYaw = 0
        cur_obj_scale(1)
    end
end

id_bhvStarSelectorStar = hook_behavior(nil, OBJ_LIST_GENACTOR, true, star_selector_star_init, star_selector_star_loop, "bhvStarSelectorStar")
id_bhvStarSelectorHandler = hook_behavior(nil, OBJ_LIST_GENACTOR, true, star_select_handler_init, star_select_handler_loop, "bhvStarSelectorHandler")

local l = gLakituState

local function mario_update(m)
    if m.playerIndex ~= 0 then return end

    if gNetworkPlayers[0].currLevelNum ~= LEVEL_ACT_SELECT then return end
    if m.action ~= ACT_DISAPPEARED then
        set_mario_action(m, ACT_DISAPPEARED, 0)
    end

    local o = obj_get_first_with_behavior_id(id_bhvStarSelectorHandler)

    obj_set_model_extended(m.marioObj, E_MODEL_NONE)
    camera_freeze()

    local baseX = o.oPosX - ((MAX_ACT_NUM - 1) * SPACING / 2)
    local middleX = baseX + ((MAX_ACT_NUM - 1) * SPACING / 2)

    l.pos.x, l.pos.y, l.pos.z = middleX, -28, 926
    l.focus.x, l.focus.y, l.focus.z = middleX, o.oPosY, o.oPosZ
end

local function for_each_object_with_behavior(behavior, func_f)
    local obj = obj_get_first_with_behavior_id(behavior)
    while obj ~= nil do
        func_f(obj)
        obj = obj_get_next_with_same_behavior_id(obj)
    end
end

local function render_challenges(x, y, act)
    local challengeNames = get_nova_star_challenge_names(currHack, currMoveset, targetWarpInfo.targetLevel, act)
    
    if challengeNames then
        djui_hud_print_text("CHALLENGES:", x, y, 1)
        for i, name in ipairs(challengeNames) do
            djui_hud_print_text("- "..name, x, y + (16 * i), 1)
        end
    end
end

local function hud_render()
    local starSelectHandler = obj_get_first_with_behavior_id(id_bhvStarSelectorHandler)

    if not starSelectHandler then return end

    djui_hud_set_resolution(RESOLUTION_N64)
    djui_hud_set_font(FONT_TINY)

    local sWidth = djui_hud_get_screen_width()
    local sHeight = djui_hud_get_screen_height()

    djui_hud_set_color(0, 0, 0, 255)

    local pos
    local out

    for_each_object_with_behavior(id_bhvStarSelectorStar, function (o)
        pos = { x = o.oPosX, y = o.oPosY, z = o.oPosZ }
        out = { x = 0, y = 0, z = 0 }

        djui_hud_world_pos_to_screen_pos(pos, out)

        djui_hud_print_text(""..o.oBehParams2ndByte + 1, out.x - 3, out.y - 35, 1)
    end)

    local selectedAct = starSelectHandler.oBehParams2ndByte

    local actName = get_nova_star_name(targetWarpInfo.targetLevel, selectedAct + 1):upper()

    local actNamePosX = (sWidth -  djui_hud_measure_text(actName)) * 0.5
    local actNamePosY = out.y + 20

    djui_hud_print_text(actName, actNamePosX, actNamePosY, 1)

    local scoreText = "MYSCORE"
    djui_hud_set_font(FONT_TINY)
    local scoreWidth = djui_hud_measure_text(scoreText)

    local coinScoreText = " $@" .. save_file_get_course_coin_score(get_current_save_file_num() - 1, targetWarpInfo.targetCourse)
    djui_hud_set_font(FONT_HUD)
    local coinScoreWidth = djui_hud_measure_text(coinScoreText)

    local totalWidth = scoreWidth + coinScoreWidth
    local centerX = (sWidth - totalWidth) * 0.5
    local scoreY = actNamePosY + 25

    djui_hud_set_font(FONT_TINY)
    djui_hud_print_text(scoreText, centerX, scoreY, 1)

    djui_hud_set_color(255, 255, 255, 255)

    djui_hud_set_font(FONT_HUD)
    djui_hud_print_text(coinScoreText, centerX + scoreWidth, scoreY, 1)

    local courseTextureTop = get_texture_info("texture_menu_course_upper")
    local courseTextureBottom = get_texture_info("texture_menu_course_lower")

    local coursePlatePosX = (sWidth * 0.5) - (courseTextureTop.width * 0.5)
    local coursePlateTopPosY = (scoreY + (courseTextureTop.height * 0.75))
    local coursePlateBottomPosY =  coursePlateTopPosY + courseTextureBottom.height

    djui_hud_set_filter(FILTER_LINEAR)
    djui_hud_render_texture(courseTextureTop, coursePlatePosX, coursePlateTopPosY, 1, 1)
    djui_hud_render_texture(courseTextureBottom, coursePlatePosX, coursePlateBottomPosY, 1, 1)
    djui_hud_set_filter(FILTER_NEAREST)

    local courseNumX = (sWidth * 0.5)
    local courseNumY = coursePlateBottomPosY - 2
    local courseNumChars = count_string_chars(""..targetWarpInfo.targetCourse)
    
    djui_hud_print_text(""..targetWarpInfo.targetCourse, courseNumX - (7 * courseNumChars), courseNumY, 1)

    djui_hud_set_font(FONT_TINY)
    djui_hud_set_color(0, 0, 0, 255)

    local levelName = get_level_name(targetWarpInfo.targetCourse, targetWarpInfo.targetLevel, targetWarpInfo.targetArea):upper()
    local levelNameWidth = djui_hud_measure_text(levelName)
    local levelNameX = (sWidth * 0.5) - (levelNameWidth * 0.5)
    local levelNameY = (coursePlateBottomPosY + courseTextureBottom.height) + 8

    djui_hud_print_text(levelName, levelNameX, levelNameY, 1)

    render_challenges(15, actNamePosY, selectedAct + 1)
end

hook_event(HOOK_ON_HUD_RENDER_BEHIND, hud_render)
hook_event(HOOK_MARIO_UPDATE, mario_update)