E_MODEL_COSMIC_CLONE = smlua_model_util_get_id("cosmic_geo")
cosmicClonePath = {}

delay = 60
cosmicClonesNum = 20

---@class Object
    ---@field oPathIndex integer
    ---@field oDelayCountdown integer
    ---@field oOwner integer
    ---@field oAnim integer
define_custom_obj_fields({
    oPathIndex = "u32",
    oDelayCountdown = "u32",
    oOwner = "u32",
    oAnim = "u32",
})

local function cosmic_clone_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    obj_set_model_extended(o, E_MODEL_COSMIC_CLONE)
    local hitbox = get_temp_object_hitbox()
    hitbox.interactType = INTERACT_DAMAGE
    hitbox.radius = 60
    hitbox.height = 160
    hitbox.damageOrCoinValue = 3
    obj_set_hitbox(o, hitbox)
    o.oPathIndex = 1
end

local cosmicCloneSpeed = 1

local function cosmic_clone_loop(o)
    if #cosmicClonePath <= 0 then return end

    local oAnimInfo = o.header.gfx.animInfo

    if not is_challenge_active(gMarioStates[0], CHALLENGE_COSMIC_CLONE) then
        obj_mark_for_deletion(o)
        spawn_mist_particles()
    end

    if o.oDelayCountdown == nil or o.oDelayCountdown <= 0 then
        local pathPoint = cosmicClonePath[o.oPathIndex]
        if pathPoint.animInfo.animID >= 0 and pathPoint.animInfo.animID ~= nil then
            o.oAnim = pathPoint.animInfo.animID
        end
        o.header.gfx.animInfo.curAnim = get_mario_vanilla_animation(o.oAnim)
        oAnimInfo.animYTrans = pathPoint.animInfo.animYTrans
        oAnimInfo.animFrameAccelAssist = pathPoint.animInfo.animFrameAccelAssist
        oAnimInfo.animAccel = pathPoint.animInfo.animAccel
        obj_set_pos(o, pathPoint.position.x, pathPoint.position.y, pathPoint.position.z)
        obj_set_face_angle(o, pathPoint.faceAngleX, pathPoint.faceAngleY, pathPoint.faceAngleZ)
        o.oPathIndex = (o.oPathIndex % #cosmicClonePath) + cosmicCloneSpeed
    else
        o.oDelayCountdown = o.oDelayCountdown - 1
    end
    o.oInteractStatus = 0
end

id_bhvCosmicClone = hook_behavior(nil, OBJ_LIST_GENACTOR, true, cosmic_clone_init, cosmic_clone_loop, "bhvCosmicClone")

local stopTimeActions = {
    [ACT_EXIT_LAND_SAVE_DIALOG] = true,
    [ACT_EXIT_AIRBORNE] = true,
}

local npl = gNetworkPlayers[0]

--* function to print a copy pastable table in console
function print_table(tableToPrint) --* chatgpt function
    local function serializeTable(tbl, indent)
        local output = "{\n"
        local indentStr = string.rep("    ", indent)

        for key, value in pairs(tbl) do
            output = output .. indentStr
            if type(key) == "number" then
                output = output .. "[" .. key .. "] = "
            elseif type(key) == "string" then
                output = output .. key .. " = "
            end

            if type(value) == "table" then
                output = output .. serializeTable(value, indent + 1)
            elseif type(value) == "string" then
                output = output .. string.format("%q", value) .. ",\n"
            else
                output = output .. tostring(value) .. ",\n"
            end
        end

        return output .. string.rep("    ", indent - 1) .. "},\n"
    end

    local result = "raceClonePath = " .. serializeTable(tableToPrint, 1)
    print(result)
end

local startRecording = false

---@param m MarioState
local function mario_update(m)
    if m.playerIndex ~= 0 then return end

    if is_challenge_active(gMarioStates[0], CHALLENGE_COSMIC_CLONE) or startRecording then

        if stopTimeActions[m.action] then
            disable_time_stop_including_mario()
        end

        table.insert(cosmicClonePath, {
            position = {x = m.pos.x, y = m.pos.y, z = m.pos.z},
            animInfo = {animID = m.marioObj.header.gfx.animInfo.animID,
                        animYTrans = m.marioObj.header.gfx.animInfo.animYTrans,
                        animFrameAccelAssist = m.marioObj.header.gfx.animInfo.animFrameAccelAssist,
                        animAccel = m.marioObj.header.gfx.animInfo.animAccel},
            faceAngleY = m.marioObj.oFaceAngleYaw,
            faceAngleX = m.marioObj.oFaceAnglePitch,
            faceAngleZ = m.marioObj.oFaceAngleRoll,
        })
    end

    if not debug or currMoveset == "omm rebirth 1.4" then return end

    --if m.controller.buttonPressed & Y_BUTTON ~= 0 then
    --    startRecording = not startRecording
    --    djui_chat_message_create(startRecording and "Recording" or "Stopped Recording")
    --elseif m.controller.buttonPressed & X_BUTTON ~= 0 then
    --    print_table(cosmicClonePath)
    --end
end

local invincActions = {
    [ACT_WARP_DOOR_SPAWN] = 70,
    [ACT_DEATH_EXIT_LAND] = 60,
    [ACT_PULLING_DOOR] = 90,
    [ACT_PUSHING_DOOR] = 90,
    [ACT_FALLING_DEATH_EXIT] = 60,
    [ACT_SPECIAL_DEATH_EXIT] = 120,
    [ACT_UNUSED_DEATH_EXIT] = 60,
    [ACT_PICKING_UP] = 15,
    [ACT_DIVE_PICKING_UP] = 30,
    [ACT_PICKING_UP_BOWSER] = 75,
    [ACT_GRAB_POLE_SLOW] = 45,
    [ACT_GRAB_POLE_FAST] = 45,
    [ACT_FIRST_PERSON] = 60,
    [ACT_SHOT_FROM_CANNON] = 30,
    [ACT_READING_NPC_DIALOG] = 75,
    [ACT_SPAWN_SPIN_LANDING] = 60,
    [ACT_SPAWN_NO_SPIN_LANDING] = 60,
    [ACT_SPAWN_SPIN_AIRBORNE] = 60,
    [ACT_EXIT_LAND_SAVE_DIALOG] = 180,
}

function on_set_mario_action(m)
    if m.playerIndex ~= 0 then return end

    if invincActions[m.action] and is_challenge_active(gMarioStates[0], CHALLENGE_COSMIC_CLONE | CHALLENGE_PVP_RACE) then
        m.invincTimer = invincActions[m.action]
    end
end

hook_event(HOOK_ON_SET_MARIO_ACTION, on_set_mario_action)
hook_event(HOOK_BEFORE_MARIO_UPDATE, mario_update)