MODEL_RAIN = smlua_model_util_get_id("rain_geo")

local RAIN_SPEED = 200

local MAX_DROPS = 100

---@param o Object
local function bhv_water_splash_init(o)
    local floor = cur_obj_update_floor_height_and_get_floor()

    local z = {x=0, y=0, z=0}
    local normal = floor and floor.normal or z

    o.oFaceAnglePitch = 16383 - calculate_pitch(z, normal)
    o.oFaceAngleYaw = calculate_yaw(z, normal)
end

hook_behavior(id_bhvWaterDropletSplash, OBJ_LIST_GENACTOR, false, bhv_water_splash_init, nil, "bhvNovaStarsWaterDropletSplash")

local function reset_rain(o)
    cur_obj_hide()
    cur_obj_set_home_once()
    cur_obj_update_floor_height_and_get_floor()
    o.oPosX = math.random(-8100, 8100)
    o.oPosY = gMarioStates[0].pos.y + math.random(2000, 10000)
    o.oPosZ = math.random(-8100, 8100)
    o.oFaceAngleYaw = gLakituState.yaw
    o.oFaceAnglePitch = 16383
    o.oFaceAngleRoll = 0
end

local function rain_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    cur_obj_hide()
    obj_set_model_extended(o, MODEL_RAIN)
    reset_rain(o)
end

---@param o Object
local function rain_loop(o)
    object_step()
    cur_obj_update_floor()

    o.oPosY = o.oPosY - RAIN_SPEED

    if not o.oFloor then
        reset_rain(o)
    else
        cur_obj_unhide()
    end

    if o.oPosY <= o.oFloorHeight + RAIN_SPEED + 10 then
        spawn_non_sync_object(id_bhvWaterDropletSplash, E_MODEL_SMALL_WATER_SPLASH, o.oPosX, o.oFloorHeight, o.oPosZ, nil)
        reset_rain(o)
    end
end

id_bhvRain = hook_behavior(nil, OBJ_LIST_LEVEL, true, rain_init, rain_loop, "bhvRain")

local function mario_update(m)
    if m.playerIndex ~= 0 then return end
    if not is_challenge_active(m, CHALLENGE_ACID_RAIN) then return end

    local rainCount = obj_count_objects_with_behavior_id(id_bhvRain)
    if rainCount < MAX_DROPS then
        spawn_non_sync_object(id_bhvRain, MODEL_RAIN, 0, m.pos.y + 1000, 0, nil)
    end
end
----- ASHES

local ashParticles = {}
local particleCount = 100
local maxZ = 3  -- "depth"

local function init_ash_particles(sWidth, sHeight)
    for i = 1, particleCount do
        ashParticles[i] = {
            x = math.random(0, 400),
            y = math.random(0, sHeight),
            z = math.random(1, maxZ),
            alive = true
        }
    end
end

local function update_ash_particles(m, sWidth, sHeight)
    local deltaX = m.vel.x * 0.05
    local deltaY = m.vel.y * 0.1
    local forwardEffect = m.forwardVel * 0.02

    for i, particle in ipairs(ashParticles) do
        if particle.alive then
            local driftY = -(1 / particle.z) - forwardEffect
            particle.x = particle.x + math.random(-1, 1) / particle.z + deltaX
            particle.y = particle.y + driftY + deltaY

            if particle.x > sWidth then
                particle.x = 0
            elseif particle.x < 0 then
                particle.x = sWidth
            end

            if particle.y < 0 then 
                particle.x = math.random(0, sWidth)
                particle.y = sHeight
                particle.z = math.random(1, maxZ)
            end
        end
    end
end

local TEXT_ASH = get_texture_info("ash")

local function hud_render_behind()
    djui_hud_set_resolution(RESOLUTION_N64)
    djui_hud_set_font(FONT_HUD)
    enableAsh = is_challenge_active(gMarioStates[0], CHALLENGE_RISING_LAVA)

    if not enableAsh then return end

    local sWidth = djui_hud_get_screen_width()
    local sHeight = djui_hud_get_screen_height()

    djui_hud_set_color(255, 255, 255, 255)

    djui_hud_set_filter(FILTER_LINEAR)

    for _, particle in ipairs(ashParticles) do
        local scale = 1 / particle.z
        djui_hud_render_texture(TEXT_ASH, particle.x, particle.y, scale, scale)
    end

    local m = gMarioStates[0]
    update_ash_particles(m, sWidth, sHeight)
end

init_ash_particles(0, djui_hud_get_screen_height())

hook_event(HOOK_ON_HUD_RENDER_BEHIND, hud_render_behind)
hook_event(HOOK_MARIO_UPDATE, mario_update)