local djui_hud_set_filter,djui_hud_set_color,djui_hud_render_texture,djui_hud_set_font,djui_hud_print_text,hud_hide,hook_event, get_texture_info =
djui_hud_set_filter,djui_hud_set_color,djui_hud_render_texture,djui_hud_set_font,djui_hud_print_text,hud_hide, hook_event, get_texture_info

local HUD_COIN_REQ = 1
local HUD_A_PRESS_COUNT = 2

local hudElementAction = {
    [HUD_COIN_REQ] =  { visible = false, xOffset = 23, yOffset = -300, targetY = 360, slideSpeed = 30, shouldSlideIn = false, shouldSlideOut = false },
    [HUD_A_PRESS_COUNT] =  { visible = false, xOffset = 23, yOffset = -300, targetY = 380, slideSpeed = 30, shouldSlideIn = false, shouldSlideOut = false },
}

function formatTimer(timer)
    local totalSeconds = timer / 30
    local minutes = math.floor(totalSeconds / 60)
    local seconds = math.floor(totalSeconds % 60)
    local milliseconds = math.floor((totalSeconds * 1000) % 1000)

    return string.format('%02d"%02d"%03d', minutes, seconds, milliseconds)
end

local function update_hud_slide(action)
    if action.shouldSlideIn and action.yOffset < action.targetY then
        action.yOffset = action.yOffset + action.slideSpeed
        if action.yOffset >= action.targetY then
            action.yOffset = action.targetY
            action.shouldSlideIn = false
        end
    end

    if action.shouldSlideOut and action.yOffset > -500 then
        action.yOffset = action.yOffset - action.slideSpeed
        if action.yOffset <= -500 then
            action.yOffset = -500
            action.visible = false
            action.shouldSlideOut = false
        end
    end
end

local function render_coin_req(x, y, scale)
    djui_hud_set_filter(FILTER_NEAREST)
    djui_hud_set_color(255, 255, 255, 255)

    djui_hud_set_font(FONT_HUD)

    local m = gMarioStates[0]

    local o = obj_get_first_with_behavior_id(id_bhvMSStar) or obj_get_first_with_behavior_id(id_bhvMSRedCoinStarSpawner) or m.marioObj

    local coins = m.numCoins
    local totalScale = scale.x + scale.y
    local averageScale = totalScale * 0.5

    djui_hud_print_text("$@"..coins.."/"..o.oCoinReq, x * averageScale, y, averageScale)
end

local function render_a_presses(x, y, scale)
    djui_hud_set_filter(FILTER_NEAREST)
    djui_hud_set_color(255, 255, 255, 255)

    djui_hud_set_font(FONT_CUSTOM_HUD)

    local m = gMarioStates[0]

    local o = obj_get_first_with_behavior_id(id_bhvMSStar) or obj_get_first_with_behavior_id(id_bhvMSRedCoinStarSpawner) or m.marioObj

    local aPresses = ps[m.playerIndex].APressCount
    local totalScale = scale.x + scale.y
    local averageScale = totalScale * 0.5

    djui_hud_print_text("A Presses : "..clamp(o.oAPressMaxCount - aPresses, 0, o.oAPressMaxCount), x * averageScale, y, averageScale)
end

local function handle_hud_slide(actionKey, condition)
    local action = hudElementAction[actionKey]
    if condition and not action.visible then
        action.visible = true
        action.shouldSlideIn = true
        action.shouldSlideOut = false
    elseif not condition and action.visible then
        action.shouldSlideOut = true
        action.shouldSlideIn = false
    end
    update_hud_slide(action)
end

local function hud_render()
    if not gGlobalSyncTable.novaStarsEnabled then return end

    djui_hud_set_resolution(RESOLUTION_DJUI)
    djui_hud_set_font(FONT_HUD)

    local m = gMarioStates[0]
    local star = obj_get_first_with_behavior_id(id_bhvMSStar) or obj_get_first_with_behavior_id(id_bhvSpawnedMSStar) or obj_get_first_with_behavior_id(id_bhvMSRedCoinStarSpawner) or m.marioObj

    handle_hud_slide(HUD_COIN_REQ, is_challenge_active(m, CHALLENGE_COIN_REQUIREMENT) and star.oCoinReq > 0)
    handle_hud_slide(HUD_A_PRESS_COUNT, is_challenge_active(m, CHALLENGE_ABC))

    local sWidth = djui_hud_get_screen_width()
    local sHeight = djui_hud_get_screen_height()

    local sHudElements = {
        [HUD_COIN_REQ] = {
            appearCondition = function() return hudElementAction[HUD_COIN_REQ].visible end,
            x = hudElementAction[HUD_COIN_REQ].xOffset,
            y = hudElementAction[HUD_COIN_REQ].yOffset,
            scale = {x = 3, y = 3},
            renderFunction = render_coin_req
        },
        [HUD_A_PRESS_COUNT] = {
            appearCondition = function() return hudElementAction[HUD_A_PRESS_COUNT].visible end,
            x = hudElementAction[HUD_A_PRESS_COUNT].xOffset,
            y = hudElementAction[HUD_A_PRESS_COUNT].yOffset,
            scale = {x = 3, y = 3},
            renderFunction = render_a_presses
        },
    }

    for _, element in pairs(sHudElements) do
        if element.appearCondition() then
            element.renderFunction(element.x, element.y, element.scale)
        end
    end

    djui_hud_set_resolution(RESOLUTION_N64)

    local bhvsToDisplay = {
        {bhv = id_bhvMSStar, texture = gTextures.star, pos = {x = -5, y = -5}, scale = {x = 0.75, y = 0.75}, color = {r = 30, g = 30, b = 255, a = 150}},
        {bhv = id_bhvSpawnedMSStar, texture = gTextures.star, pos = {x = -5, y = -5}, scale = {x = 0.75, y = 0.75}, color = {r = 30, g = 30, b = 255, a = 150}},
        {bhv = id_bhvMSRedCoin, texture = gTextures.coin, pos = {x = -5, y = -5}, scale = {x = 0.75, y = 0.75}, color = {r = 0, g = 255, b = 0, a = 150}},
        {bhv = id_bhvNovaPipe, texture = TEXTURE_NOVA_PIPE, pos = {x = -5, y = -5}, scale = {x = 0.75, y = 0.75}, color = {r = 255, g = 255, b = 255, a = 150}},
    }
    
    for i, entry in pairs(bhvsToDisplay) do
        local pos
        local out

        for_each_object_with_behavior(entry.bhv, function (o)
            pos = { x = o.oPosX, y = o.oPosY, z = o.oPosZ }
            out = { x = 0, y = 0, z = 0 }
    
            djui_hud_world_pos_to_screen_pos(pos, out)
    
            out.x = clamp(out.x, 0, djui_hud_get_screen_width())
            out.y = clamp(out.y, 0, djui_hud_get_screen_height())

            djui_hud_set_color(entry.color.r, entry.color.g, entry.color.b, entry.color.a)
    
            djui_hud_render_texture(entry.texture, out.x + entry.pos.x, out.y + entry.pos.y, entry.scale.x, entry.scale.y)
        end)
    end

    for_each_object_with_behavior(id_bhvNovaPipe, function (o)
        local pipePos
        local pipeOut

        if o.oAction == 2 then

            pipePos = { x = o.oPosX, y = o.oPosY + 300, z = o.oPosZ }
            pipeOut = { x = 0, y = 0, z = 0 }

            djui_hud_world_pos_to_screen_pos(pipePos, pipeOut)

            pipeOut.x = clamp(pipeOut.x, 0, djui_hud_get_screen_width())
            pipeOut.y = clamp(pipeOut.y, 0, djui_hud_get_screen_height())

            djui_hud_set_color(0, 0, 0, 255)

            djui_hud_set_font(FONT_TINY)
            local levelName = get_level_name(o.oBehParams2ndByte, o.oBehParams & 0xFF, 1)
            local levelNameLength = djui_hud_measure_text(get_level_name(o.oBehParams2ndByte, o.oBehParams >> 24, 1))
            djui_hud_print_text(levelName, (pipeOut.x - (levelNameLength * 0.5)) + 1, pipeOut.y + 1, 1)
            djui_hud_set_color(255, 255, 255, 255)
            djui_hud_print_text(levelName, pipeOut.x - (levelNameLength * 0.5), pipeOut.y, 1)
            djui_hud_set_font(FONT_HUD)
        end
    end)

    if is_challenge_active(m, CHALLENGE_DAREDEVIL) and currMoveset == "vanilla moveset" then
        hud_set_value(HUD_DISPLAY_FLAGS, hud_get_value(HUD_DISPLAY_FLAGS) & ~HUD_DISPLAY_FLAGS_POWER)
        hud_set_value(HUD_DISPLAY_FLAGS, hud_get_value(HUD_DISPLAY_FLAGS) & ~HUD_DISPLAY_FLAG_POWER)

        local healthX = 22
        local healthY = 25

        djui_hud_set_color(0, 0, 0, 255)

        djui_hud_render_texture(TEXTURE_DAREDEVIL_OVERLAY, 0, 0, sWidth / (TEXTURE_DAREDEVIL_OVERLAY.width * 4), sHeight / (TEXTURE_DAREDEVIL_OVERLAY.height * 4))

        djui_hud_render_texture(TEXTURE_DAREDEVIL_DEAD, healthX + 1, healthY + 1, 1, 1)
        djui_hud_render_texture(TEXTURE_DAREDEVIL_DEAD, healthX - 1, healthY + 1, 1, 1)
        djui_hud_render_texture(TEXTURE_DAREDEVIL_DEAD, healthX + 1, healthY, 1, 1)
        djui_hud_render_texture(TEXTURE_DAREDEVIL_DEAD, healthX - 1, healthY, 1, 1)

        djui_hud_set_color(255, 255, 255, 255)

        if m.health > 0xFF then
            djui_hud_render_texture(TEXTURE_DAREDEVIL, healthX, healthY, 1, 1)
        else
            djui_hud_render_texture(TEXTURE_DAREDEVIL_DEAD, healthX, healthY, 1, 1)
        end
    end

    m.numLives = 99
    hud_set_value(HUD_DISPLAY_FLAGS, hud_get_value(HUD_DISPLAY_FLAGS) & ~HUD_DISPLAY_FLAG_LIVES)

    djui_hud_set_resolution(RESOLUTION_DJUI)
end

hook_event(HOOK_ON_HUD_RENDER_BEHIND, hud_render)