
local function lava_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    o.header.gfx.skipInViewCheck = true
    o.collisionData = COL_LAVA
    o.oCollisionDistance = 32767
end

local function lava_loop(o)
    load_object_collision_model()
    o.oPosY = o.oPosY + o.oBehParams2ndByte
    o.oFaceAngleYaw = o.oFaceAngleYaw + 0x10
end

id_bhvLava = hook_behavior(nil, OBJ_LIST_SURFACE, true, lava_init, lava_loop, "bhvLava")

local function quicksand_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    o.header.gfx.skipInViewCheck = true
    o.collisionData = COL_QUICKSAND
    o.oCollisionDistance = 32767
end

---@param o Object
local function quicksand_loop(o)
    load_object_collision_model()
    ---@type MarioState
    local m = gMarioStates[0]
    if o.oPosY < m.pos.y - o.oBehParams then
        o.oPosY = approach_f32_symmetric(o.oPosY, m.pos.y - o.oBehParams, 50)
    end
    o.oFaceAngleYaw = o.oFaceAngleYaw + 0x10
end

id_bhvQuicksand = hook_behavior(nil, OBJ_LIST_SURFACE, true, quicksand_init, quicksand_loop, "bhvQuicksand")

local function whomp_loop(o) --* this is pretty bad
    if np[0].currActNum < 7 then return end
    o.oNumLootCoins = 0
    if o.oAction == 8 and o.oBehParams2ndByte == 0 and o.oTimer == 0 then
        spawn_sync_object(id_bhvMrIBlueCoin, E_MODEL_BLUE_COIN, o.oPosX, o.oPosY, o.oPosZ, function (coin)
            coin.oVelY = 20
        end)
    end
end

hook_behavior(id_bhvSmallWhomp, OBJ_LIST_SURFACE, false, nil, whomp_loop, "bhvNovaWhomp")

local function lamp_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    o.oPosY = find_floor_height(o.oPosX, o.oPosY, o.oPosZ)
end

local function lamp_loop(o)

end

id_bhvLamp = hook_behavior(nil, OBJ_LIST_GENACTOR, true, lamp_init, lamp_loop, "bhvLamp")

local function star_shine_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    o.oFaceAnglePitch = 0
end

local function star_shine_loop(o)
    o.oPosY = o.parentObj.oPosY
    o.oPosX = o.parentObj.oPosX
    o.oPosZ = o.parentObj.oPosZ
    o.oFaceAngleYaw = gLakituState.yaw

    o.activeFlags = o.parentObj.activeFlags

    if o.parentObj.oInteractStatus == INT_STATUS_INTERACTED then
        obj_mark_for_deletion(o)
    end

    if o.oAction == 0 then
        local waveAmplitude = 1
        local initialScale = o.oF32Value
        local waveFrequency = 0.15

        local horizontalWaveAngle = (o.oTimer) * waveFrequency
        local horizontalScale = initialScale + math.sin(horizontalWaveAngle) * waveAmplitude

        local verticalWaveAngle = (o.oTimer / 32768) * math.pi
        local verticalScale = initialScale + math.sin(verticalWaveAngle) * waveAmplitude

        verticalScale = clampf(verticalScale, initialScale + waveAmplitude, initialScale - waveAmplitude)

        obj_scale_xyz(o, horizontalScale, verticalScale, 1)

        if horizontalScale <= initialScale * 0.05 then
            o.oAction = 1
            o.oTimer = 0
            cur_obj_hide()
        end
    elseif o.oAction == 1 then
        o.oAction = 0
        o.oFaceAngleRoll = math.random(-32767, 32767)
        cur_obj_unhide()
    end
end

id_bhvStarShine = hook_behavior(nil, OBJ_LIST_GENACTOR, true, star_shine_init, star_shine_loop, "bhvNovaStarShine")

local startMarioPosInfo = {
    action = ACT_FREEFALL,
    freeze = 0,
    pos = {x = 0, y = 0, z = 0},
    vel = {x = 0, y = 0, z = 0},
    angle = {x = 0, y = 0, z = 0}
}

local function racer_init(o)
    local act = ((gNetworkPlayers[0].currActNum - (gNetworkPlayers[0].globalIndex + 1)) % 6) + 1
    local info = gRacePaths[currHack][currMoveset][gNetworkPlayers[0].currLevelNum][gNetworkPlayers[0].currAreaIndex][act]
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    obj_set_model_extended(o, E_MODEL_COSMIC_CLONE)
    o.oPathIndex = 1

    o.oPosX = info.idlePos.x
    o.oPosY = info.idlePos.y
    o.oPosZ = info.idlePos.z
    o.oFaceAngleYaw = info.idlePos.yaw
    o.oMoveAngleYaw = info.idlePos.yaw

    local m = gMarioStates[0]

    vec3f_copy(startMarioPosInfo.pos, m.pos)
    vec3f_copy(startMarioPosInfo.vel, m.vel)
    vec3f_copy(startMarioPosInfo.angle, m.faceAngle)

    startMarioPosInfo.action = m.action
    startMarioPosInfo.actionArg = m.actionArg
    startMarioPosInfo.freeze = m.freeze

    o.parentObj = obj_get_first_with_behavior_id(id_bhvMSStar) or obj_get_first_with_behavior_id(id_bhvSpawnedMSStar)
end

local transitionLength = 30
local function racer_loop(o)
    local np = gNetworkPlayers[0]
    local m = gMarioStates[0]

    local act = ((np.currActNum - (np.globalIndex + 1)) % 6) + 1
    local info = gRacePaths[currHack][currMoveset][np.currLevelNum][np.currAreaIndex][act]

    if not is_challenge_active(m, CHALLENGE_PVP_RACE) then
        obj_mark_for_deletion(o)
        spawn_mist_particles()
        return
    end

    if o.oAction == 0 then
        cur_obj_push_mario_away_from_cylinder(100, 60)
        o.header.gfx.animInfo.curAnim = get_mario_vanilla_animation(MARIO_ANIM_FIRST_PERSON)

        if dist_between_objects(o, m.marioObj) < 200 then
            m.freeze = 1
            if cutscene_object_with_dialog(CUTSCENE_DIALOG, o, DIALOG_005) ~= 0 then
                o.oAction = 1
                play_transition(WARP_TRANSITION_FADE_INTO_CIRCLE, transitionLength, 0, 0, 0)
                o.oTimer = 0
            end
        end

        set_time_stop_flags(TIME_STOP_ENABLED)
        o.activeFlags = o.activeFlags | ACTIVE_FLAG_INITIATED_TIME_STOP

    elseif o.oAction == 1 then
        m.freeze = 1
        if o.oTimer > transitionLength then
            o.oAction = 2
            disable_time_stop_including_mario()

            vec3f_copy(m.pos, startMarioPosInfo.pos)
            vec3f_copy(m.faceAngle, startMarioPosInfo.angle)
            vec3f_copy(m.vel, startMarioPosInfo.vel)
            set_mario_action(m, startMarioPosInfo.action, startMarioPosInfo.actionArg)
            m.freeze = startMarioPosInfo.freeze

            play_transition(WARP_TRANSITION_FADE_FROM_CIRCLE, 10, 0, 0, 0)
        end

    elseif o.oAction == 2 then
        local pathPoint = info.path[o.oPathIndex]
        local animInfo = pathPoint.animInfo
        local oAnimInfo = o.header.gfx.animInfo

        if animInfo.animID and animInfo.animID >= 0 then
            o.oAnim = animInfo.animID
            oAnimInfo.curAnim = get_mario_vanilla_animation(o.oAnim)
            oAnimInfo.animYTrans = animInfo.animYTrans
            oAnimInfo.animFrameAccelAssist = animInfo.animFrameAccelAssist
            oAnimInfo.animAccel = animInfo.animAccel
        end

        obj_set_pos(o, pathPoint.position.x, pathPoint.position.y, pathPoint.position.z)
        obj_set_face_angle(o, pathPoint.faceAngleX, pathPoint.faceAngleY, pathPoint.faceAngleZ)

        o.oPathIndex = o.oPathIndex + 1
        if o.oPathIndex > #info.path then
            o.oAction = 3
            play_sound(SOUND_GENERAL_FLAME_OUT, gGlobalSoundSource)
        end

    elseif o.oAction == 3 then
        o.oPosX = o.parentObj.oPosX
        o.oPosY = o.parentObj.oPosY + 110
        o.oPosZ = o.parentObj.oPosZ
        o.oFaceAngleYaw = o.parentObj.oFaceAngleYaw

        o.header.gfx.animInfo.curAnim = get_mario_vanilla_animation(MARIO_ANIM_HANDSTAND_IDLE)
    end
end

id_bhvRacer = hook_behavior(nil, OBJ_LIST_GENACTOR, true, racer_init, racer_loop, "bhvNovaRacer")

local function celebration_star_set_model(o)

    local m = gMarioStates[0]

    if obj_has_model_extended(m.interactObj, m.interactObj.oModel) then
        obj_set_model_extended(o, m.interactObj.oModel)
    elseif obj_has_model_extended(m.interactObj, gNovaStarModels[currHack][currMoveset].transparent) then
        obj_set_model_extended(o, gNovaStarModels[currHack][currMoveset].transparent)
    end
end

hook_behavior(id_bhvCelebrationStar, OBJ_LIST_LEVEL, false, nil, celebration_star_set_model, "bhvNovaCelebrationStar")

local function falling_nova_star_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_MOVE_XZ_USING_FVEL
    o.oGravity = -2

    local customModelInfo = gNovaStarModels[currHack][currMoveset]
    if customModelInfo.shine then
        for i = 0, 3 do
            spawn_non_sync_object(id_bhvStarShine, customModelInfo.shine, 0, 0, 0, function (shine)
                shine.parentObj = o
                shine.oFaceAngleRoll = math.random(-32767, 32767) * i
                shine.oF32Value = math.random() * 0.25
                shine.oBehParams2ndByte = i
            end)
        end
    end
end

local function falling_nova_star_loop(o)
    o.oFaceAngleYaw = o.oFaceAngleYaw + math.random(0, 0x800)
    o.oFaceAnglePitch = o.oFaceAngleYaw  + math.random(0, 0x800)
    o.oFaceAngleRoll = o.oFaceAngleYaw + math.random(0, 0x800)
    o.oVelY = o.oVelY + o.oGravity
    o.oPosY = o.oPosY + o.oVelY
    if o.oTimer >= 300 then
        obj_mark_for_deletion(o)
    end
end

id_bhvFallingNovaStar = hook_behavior(nil, OBJ_LIST_GENACTOR, true, falling_nova_star_init, falling_nova_star_loop, "bhvFallingNovaStar")

local function grand_star_loop(o)
    if o.oAction == 2 then

        cur_obj_become_intangible()
        if o.oTimer >= 30 and o.oTimer <= 150 then
            cur_obj_shake_y(9)
            cur_obj_play_sound_1(SOUND_AIR_BOBOMB_LIT_FUSE)
        elseif o.oTimer > 150 then
            cur_obj_play_sound_1(SOUND_GENERAL_BREAK_BOX)
            spawn_mist_particles()
            spawn_triangle_break_particles(20, 138, 0.7, 3)
            play_puzzle_jingle()
            gGlobalSyncTable.novaStarsEnabled = true
            obj_mark_for_deletion(o)
            for i = 1, 100 do
                spawn_non_sync_object(id_bhvFallingNovaStar, gNovaStarModels[currHack][currMoveset].normal, o.oPosX, o.oPosY, o.oPosZ, function (obj)
                    obj.oMoveAngleYaw = math.random(0, 65535)
                    obj.oForwardVel = math.random(0, 120)
                    obj.oVelY = math.random(0, 150)
                    obj.oBehParams2ndByte = i
                end)
            end
        end
    end
end

--hook_behavior(id_bhvGrandStar, OBJ_LIST_LEVEL, false, nil, grand_star_loop, 'bhvNovaGrandStar')