local ACTION_IDLE = 0
local ACTION_SCALE_UP = 1
local ACTION_READY = 2
local ACTION_SCALE_DOWN = 3

local ENTER_DISTANCE = 700
local EXIT_DISTANCE = 1000
local SCALE_DURATION = 15
local SCALE_DOWN_DURATION = 15
local TRANSITION_LENGTH = 15

local function nova_pipe_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    o.collisionData = gGlobalObjectCollisionData.warp_pipe_seg3_collision_03009AC8
    o.oPosY = o.oPosY - 1
    local hitbox = get_temp_object_hitbox()
    hitbox.interactType = INTERACT_WATER_RING
    hitbox.radius = 60
    hitbox.height = 60

    o.oAction = 0
    o.oFaceAnglePitch = 0
    o.oFaceAngleRoll = 0
    
    cur_obj_hide()
    obj_set_model_extended(o, E_MODEL_NOVA_PIPE)
    obj_set_hitbox(o, hitbox)
end

local function nova_pipe_loop(o)
    load_object_collision_model()

    local m = nearest_mario_state_to_object(o)
    local dist_from_mario = dist_between_objects(o, m.marioObj)

    if o.oAction == ACTION_IDLE then
        if dist_from_mario < ENTER_DISTANCE then
            o.oAction = ACTION_SCALE_UP
            play_sound(SOUND_MENU_EXIT_PIPE, m.marioObj.header.gfx.cameraToObject)
            o.oTimer = 0
        end

    elseif o.oAction == ACTION_SCALE_UP then
        cur_obj_scale_over_time(2, SCALE_DURATION, 0.02, 1)
        cur_obj_unhide()
        if o.oTimer >= SCALE_DURATION then
            o.oAction = ACTION_READY
            o.oTimer = 0
        end

    elseif o.oAction == ACTION_READY then
        cur_obj_become_tangible()

        if o.oTimer <= 1 then
            o.oTimer = 0
        end

        if o.oTimer > TRANSITION_LENGTH then
            targetWarpInfo.targetCourse = o.oBehParams2ndByte
            targetWarpInfo.targetLevel = o.oBehParams & 0xFF
            MAX_ACT_NUM = get_level_nova_star_count(targetWarpInfo.targetLevel)
            warp_to_level(LEVEL_ACT_SELECT, 1, network_global_index_from_local(m.playerIndex))
        end

        if o.oInteractStatus == INT_STATUS_INTERACTED then
            play_sound(SOUND_MENU_ENTER_HOLE, gGlobalSoundSource)
            set_mario_action(m, ACT_DISAPPEARED, 0)
            play_transition(WARP_TRANSITION_FADE_INTO_CIRCLE, TRANSITION_LENGTH, 0, 0, 0)
            o.oTimer = 2
        end

        if dist_from_mario > EXIT_DISTANCE then
            o.oAction = ACTION_SCALE_DOWN
            cur_obj_become_intangible()
            play_sound(SOUND_MENU_ENTER_PIPE, m.marioObj.header.gfx.cameraToObject)
            o.oTimer = 0
        end

    elseif o.oAction == ACTION_SCALE_DOWN then
        cur_obj_scale_over_time(2, SCALE_DOWN_DURATION, 1, 0)
        if o.oTimer >= SCALE_DOWN_DURATION then
            o.oAction = ACTION_IDLE
        end
    end

    o.oInteractStatus = 0
end

id_bhvNovaPipe = hook_behavior(nil, OBJ_LIST_SURFACE, true, nova_pipe_init, nova_pipe_loop, "bhvNovaPipe")